# Copyright 2021 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Optional, Sequence, Tuple, Union, cast

import gpflow
import tensorflow as tf
import tensorflow_probability as tfp
from check_shapes import inherit_check_shapes
from gpflow.conditionals.util import sample_mvn
from gpflow.inducing_variables import (
    SeparateIndependentInducingVariables,
    SharedIndependentInducingVariables,
)
from gpflow.keras import tf_keras
from gpflow.logdensities import multivariate_normal
from gpflow.models import GPR, SGPR, SVGP, VGP
from gpflow.models.vgp import update_vgp_data
from gpflow.utilities import add_noise_cov, is_variable, multiple_assign, read_values
from gpflow.utilities.ops import leading_transpose

from ...data import (
    Dataset,
    add_fidelity_column,
    check_and_extract_fidelity_query_points,
    split_dataset_by_fidelity,
)
from ...space import EncoderFunction
from ...types import TensorType
from ...utils import DEFAULTS, jit
from ...utils.misc import flatten_leading_dims
from ..interfaces import (
    EncodedFastUpdateModel,
    HasTrajectorySampler,
    SupportsCovarianceWithTopFidelity,
    SupportsGetInducingVariables,
    SupportsGetInternalData,
    SupportsPredictY,
    TrainableProbabilisticModel,
    TrajectorySampler,
)
from ..optimizer import BatchOptimizer, Optimizer, OptimizeResult
from .inducing_point_selectors import InducingPointSelector
from .interface import EncodedSupportsCovarianceBetweenPoints, GPflowPredictor
from .sampler import DecoupledTrajectorySampler, RandomFourierFeatureTrajectorySampler
from .utils import (
    _covariance_between_points_for_variational_models,
    _whiten_points,
    assert_data_is_compatible,
    check_optimizer,
    randomize_hyperparameters,
    squeeze_hyperparameters,
)


class GaussianProcessRegression(
    GPflowPredictor,
    EncodedFastUpdateModel,
    EncodedSupportsCovarianceBetweenPoints,
    SupportsGetInternalData,
    HasTrajectorySampler,
):
    """
    A :class:`TrainableProbabilisticModel` wrapper for a GPflow :class:`~gpflow.models.GPR`.

    As Bayesian optimization requires a large number of sequential predictions (i.e. when maximizing
    acquisition functions), rather than calling the model directly at prediction time we instead
    call the posterior objects built by these models. These posterior objects store the
    pre-computed Gram matrices, which can be reused to allow faster subsequent predictions. However,
    note that these posterior objects need to be updated whenever the underlying model is changed
    by calling :meth:`update_posterior_cache` (this
    happens automatically after calls to :meth:`update` or :math:`optimize`).
    """

    def __init__(
        self,
        model: GPR,
        optimizer: Optimizer | None = None,
        num_kernel_samples: int = 10,
        num_rff_features: int = 1000,
        use_decoupled_sampler: bool = True,
        encoder: EncoderFunction | None = None,
    ):
        """
        :param model: The GPflow model to wrap.
        :param optimizer: The optimizer with which to train the model. Defaults to
            :class:`~trieste.models.optimizer.Optimizer` with :class:`~gpflow.optimizers.Scipy`.
        :param num_kernel_samples: Number of randomly sampled kernels (for each kernel parameter) to
            evaluate before beginning model optimization. Therefore, for a kernel with `p`
            (vector-valued) parameters, we evaluate `p * num_kernel_samples` kernels.
        :param num_rff_features: The number of random Fourier features used to approximate the
            kernel when calling :meth:`trajectory_sampler`. We use a default of 1000 as it
            typically perfoms well for a wide range of kernels. Note that very smooth
            kernels (e.g. RBF) can be well-approximated with fewer features.
        :param use_decoupled_sampler: If True use a decoupled random Fourier feature sampler, else
            just use a random Fourier feature sampler. The decoupled sampler suffers less from
            overestimating variance and can typically get away with a lower num_rff_features.
        :param encoder: Optional encoder with which to transform query points before
            generating predictions.
        """
        super().__init__(optimizer, encoder)
        self._model = model

        check_optimizer(self.optimizer)

        if num_kernel_samples < 0:
            raise ValueError(
                f"num_kernel_samples must be greater or equal to zero but got {num_kernel_samples}."
            )
        self._num_kernel_samples = num_kernel_samples

        if num_rff_features <= 0:
            raise ValueError(
                f"num_rff_features must be greater than zero but got {num_rff_features}."
            )
        self._num_rff_features = num_rff_features
        self._use_decoupled_sampler = use_decoupled_sampler
        self._ensure_variable_model_data()
        self.create_posterior_cache()

    def __repr__(self) -> str:
        """"""
        return (
            f"GaussianProcessRegression({self.model!r}, {self.optimizer!r},"
            f"{self._num_kernel_samples!r}, {self._num_rff_features!r},"
            f"{self._use_decoupled_sampler!r})"
        )

    @property
    def model(self) -> GPR:
        return self._model

    def _ensure_variable_model_data(self) -> None:
        # GPflow stores the data in Tensors. However, since we want to be able to update the data
        # without having to retrace the acquisition functions, put it in Variables instead.
        # Data has to be stored in variables with dynamic shape to allow for changes
        # Sometimes, for instance after serialization-deserialization, the shape can be overridden
        # Thus here we ensure data is stored in dynamic shape Variables

        if all(is_variable(x) and x.shape[0] is None for x in self._model.data):
            # both query points and observations are in right shape
            # nothing to do
            return

        self._model.data = (
            tf.Variable(
                self._model.data[0], trainable=False, shape=[None, *self._model.data[0].shape[1:]]
            ),
            tf.Variable(
                self._model.data[1], trainable=False, shape=[None, *self._model.data[1].shape[1:]]
            ),
        )

    def predict_y_encoded(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
        f_mean, f_var = self.predict_encoded(query_points)
        return self.model.likelihood.predict_mean_and_var(query_points, f_mean, f_var)

    def update_encoded(self, dataset: Dataset) -> None:
        self._ensure_variable_model_data()

        x, y = self.model.data[0].value(), self.model.data[1].value()

        assert_data_is_compatible(dataset, Dataset(x, y))

        if dataset.query_points.shape[-1] != x.shape[-1]:
            raise ValueError

        if dataset.observations.shape[-1] != y.shape[-1]:
            raise ValueError

        self.model.data[0].assign(dataset.query_points)
        self.model.data[1].assign(dataset.observations)
        self.update_posterior_cache()

    def covariance_between_points_encoded(
        self, query_points_1: TensorType, query_points_2: TensorType
    ) -> TensorType:
        r"""
        Compute the posterior covariance between sets of query points.

        .. math:: \Sigma_{12} = K_{12} - K_{x1}(K_{xx} + \sigma^2 I)^{-1}K_{x2}

        Note that query_points_2 must be a rank 2 tensor, but query_points_1 can
        have leading dimensions.

        :param query_points_1: Set of query points with shape [..., N, D]
        :param query_points_2: Sets of query points with shape [M, D]
        :return: Covariance matrix between the sets of query points with shape [..., L, N, M]
            (L being the number of latent GPs = number of output dimensions)
        """
        tf.debugging.assert_shapes(
            [(query_points_1, [..., "N", "D"]), (query_points_2, ["M", "D"])]
        )

        x = self.model.data[0].value()
        num_data = tf.shape(x)[0]
        s = tf.linalg.diag(tf.fill([num_data], self.model.likelihood.variance))

        K = self.model.kernel(x)  # [num_data, num_data] or [L, num_data, num_data]
        Kx1 = self.model.kernel(query_points_1, x)  # [..., N, num_data] or [..., L, N, num_data]
        Kx2 = self.model.kernel(x, query_points_2)  # [num_data, M] or [L, num_data, M]
        K12 = self.model.kernel(query_points_1, query_points_2)  # [..., N, M] or [..., L, N, M]

        if len(tf.shape(K)) == 2:
            # if single output GPR, the kernel does not return the latent dimension so
            # we add it back here
            K = tf.expand_dims(K, -3)
            Kx1 = tf.expand_dims(Kx1, -3)
            Kx2 = tf.expand_dims(Kx2, -3)
            K12 = tf.expand_dims(K12, -3)
        elif len(tf.shape(K)) > 3:
            raise NotImplementedError(
                "Covariance between points is not supported "
                "for kernels of type "
                f"{type(self.model.kernel)}."
            )

        L = tf.linalg.cholesky(K + s)  # [L, num_data, num_data]

        Kx1 = leading_transpose(Kx1, [..., -1, -2])  # [..., L, num_data, N]
        Linv_Kx1 = tf.linalg.triangular_solve(L, Kx1)  # [..., L, num_data, N]
        Linv_Kx2 = tf.linalg.triangular_solve(L, Kx2)  # [L, num_data, M]

        # The line below is just A^T*B over the last 2 dimensions.
        cov = K12 - tf.einsum("...lji,ljk->...lik", Linv_Kx1, Linv_Kx2)  # [..., L, N, M]

        num_latent = self.model.num_latent_gps
        if cov.shape[-3] == 1 and num_latent > 1:
            # For multioutput GPR with shared kernel, we need to duplicate cov
            # for each output
            cov = tf.repeat(cov, num_latent, axis=-3)

        tf.debugging.assert_shapes(
            [
                (query_points_1, [..., "N", "D"]),
                (query_points_2, ["M", "D"]),
                (cov, [..., "L", "N", "M"]),
            ]
        )

        return cov

    def optimize_encoded(self, dataset: Dataset) -> OptimizeResult:
        """
        Optimize the model with the specified `dataset`.

        For :class:`GaussianProcessRegression`, we (optionally) try multiple randomly sampled
        kernel parameter configurations as well as the configuration specified when initializing
        the kernel. The best configuration is used as the starting point for model optimization.

        For trainable parameters constrained to lie in a finite interval (through a sigmoid
        bijector), we begin model optimization from the best of a random sample from these
        parameters' acceptable domains.

        For trainable parameters without constraints but with priors, we begin model optimization
        from the best of a random sample from these parameters' priors.

        For trainable parameters with neither priors nor constraints, we begin optimization from
        their initial values.

        :param dataset: The data with which to optimize the `model`.
        """
        num_trainable_params_with_priors_or_constraints = tf.reduce_sum(
            [
                tf.size(param)
                for param in self.model.trainable_parameters
                if param.prior is not None or isinstance(param.bijector, tfp.bijectors.Sigmoid)
            ]
        )

        if (
            min(num_trainable_params_with_priors_or_constraints, self._num_kernel_samples) >= 1
        ):  # Find a promising kernel initialization
            self.find_best_model_initialization(
                self._num_kernel_samples * num_trainable_params_with_priors_or_constraints
            )
        result = self.optimizer.optimize(self.model, dataset)
        self.update_posterior_cache()
        return result

    def find_best_model_initialization(self, num_kernel_samples: int) -> None:
        """
        Test `num_kernel_samples` models with sampled kernel parameters. The model's kernel
        parameters are then set to the sample achieving maximal likelihood.

        :param num_kernel_samples: Number of randomly sampled kernels to evaluate.
        """

        @tf.function
        def evaluate_loss_of_model_parameters() -> tf.Tensor:
            randomize_hyperparameters(self.model)
            return self.model.training_loss()

        squeeze_hyperparameters(self.model)
        current_best_parameters = read_values(self.model)
        min_loss = self.model.training_loss()

        for _ in tf.range(num_kernel_samples):
            try:
                train_loss = evaluate_loss_of_model_parameters()
            except tf.errors.InvalidArgumentError:  # allow badly specified kernel params
                train_loss = 1e100

            if train_loss < min_loss:  # only keep best kernel params
                min_loss = train_loss
                current_best_parameters = read_values(self.model)

        multiple_assign(self.model, current_best_parameters)

    def trajectory_sampler(self) -> TrajectorySampler[GaussianProcessRegression]:
        """
        Return a trajectory sampler. For :class:`GaussianProcessRegression`, we build
        trajectories using a random Fourier feature approximation.

        At the moment only models with single latent GP are supported.

        :return: The trajectory sampler.
        :raise NotImplementedError: If we try to use the
            sampler with a model that has more than one latent GP.
        """
        if self.model.num_latent_gps > 1:
            raise NotImplementedError(
                f"""
                Trajectory sampler does not currently support models with multiple latent
                GPs, however received a model with {self.model.num_latent_gps} latent GPs.
                """
            )

        if self._use_decoupled_sampler:
            return DecoupledTrajectorySampler(self, self._num_rff_features)
        else:
            return RandomFourierFeatureTrajectorySampler(self, self._num_rff_features)

    def get_internal_data(self) -> Dataset:
        """
        Return the model's training data.

        :return: The model's training data.
        """
        return Dataset(self.model.data[0], self.model.data[1])

    def conditional_predict_f_encoded(
        self, query_points: TensorType, additional_data: Dataset
    ) -> tuple[TensorType, TensorType]:
        """
        Returns the marginal GP distribution at query_points conditioned on both the model
        and some additional data, using exact formula. See :cite:`chevalier2014corrected`
        (eqs. 8-10) for details.

        :param query_points: Set of query points with shape [M, D]
        :param additional_data: Dataset with query_points with shape [..., N, D] and observations
                 with shape [..., N, L]
        :return: mean_qp_new: predictive mean at query_points, with shape [..., M, L],
                 and var_qp_new: predictive variance at query_points, with shape [..., M, L]
        """
        tf.debugging.assert_shapes(
            [
                (additional_data.query_points, [..., "N", "D"]),
                (additional_data.observations, [..., "N", "L"]),
                (query_points, ["M", "D"]),
            ],
            message="additional_data must have query_points with shape [..., N, D]"
            " and observations with shape [..., N, L], and query_points "
            "should have shape [M, D]",
        )

        mean_add, cov_add = self.predict_joint_encoded(
            additional_data.query_points
        )  # [..., N, L], [..., L, N, N]
        mean_qp, var_qp = self.predict_encoded(query_points)  # [M, L], [M, L]

        cov_cross = self.covariance_between_points(
            additional_data.query_points, query_points
        )  # [..., L, N, M]

        cov_shape = tf.shape(cov_add)
        noise = self.get_observation_noise() * tf.eye(
            cov_shape[-2], batch_shape=cov_shape[:-2], dtype=cov_add.dtype
        )
        L_add = tf.linalg.cholesky(cov_add + noise)  # [..., L, N, N]
        A = tf.linalg.triangular_solve(L_add, cov_cross, lower=True)  # [..., L, N, M]
        var_qp_new = var_qp - leading_transpose(
            tf.reduce_sum(A**2, axis=-2), [..., -1, -2]
        )  # [..., M, L]

        mean_add_diff = additional_data.observations - mean_add  # [..., N, L]
        mean_add_diff = leading_transpose(mean_add_diff, [..., -1, -2])[..., None]  # [..., L, N, 1]
        AM = tf.linalg.triangular_solve(L_add, mean_add_diff)  # [..., L, N, 1]

        mean_qp_new = mean_qp + leading_transpose(
            (tf.matmul(A, AM, transpose_a=True)[..., 0]), [..., -1, -2]
        )  # [..., M, L]

        tf.debugging.assert_shapes(
            [
                (additional_data.observations, [..., "N", "L"]),
                (query_points, ["M", "D"]),
                (mean_qp_new, [..., "M", "L"]),
                (var_qp_new, [..., "M", "L"]),
            ],
            message="received unexpected shapes computing conditional_predict_f,"
            "check model kernel structure?",
        )

        return mean_qp_new, var_qp_new

    def conditional_predict_joint_encoded(
        self, query_points: TensorType, additional_data: Dataset
    ) -> tuple[TensorType, TensorType]:
        """
        Predicts the joint GP distribution at query_points conditioned on both the model
        and some additional data, using exact formula. See :cite:`chevalier2014corrected`
        (eqs. 8-10) for details.

        :param query_points: Set of query points with shape [M, D]
        :param additional_data: Dataset with query_points with shape [..., N, D] and observations
                 with shape [..., N, L]
        :return: mean_qp_new: predictive mean at query_points, with shape [..., M, L],
                 and cov_qp_new: predictive covariance between query_points, with shape
                 [..., L, M, M]
        """
        tf.debugging.assert_shapes(
            [
                (additional_data.query_points, [..., "N", "D"]),
                (additional_data.observations, [..., "N", "L"]),
                (query_points, ["M", "D"]),
            ],
            message="additional_data must have query_points with shape [..., N, D]"
            " and observations with shape [..., N, L], and query_points "
            "should have shape [M, D]",
        )

        leading_dims = tf.shape(additional_data.query_points)[:-2]  # [...]
        new_shape = tf.concat([leading_dims, tf.shape(query_points)], axis=0)  # [..., M, D]
        query_points_r = tf.broadcast_to(query_points, new_shape)  # [..., M, D]
        points = tf.concat([additional_data.query_points, query_points_r], axis=-2)  # [..., N+M, D]

        mean, cov = self.predict_joint_encoded(points)  # [..., N+M, L], [..., L, N+M, N+M]

        N = tf.shape(additional_data.query_points)[-2]

        mean_add = mean[..., :N, :]  # [..., N, L]
        mean_qp = mean[..., N:, :]  # [..., M, L]

        cov_add = cov[..., :N, :N]  # [..., L, N, N]
        cov_qp = cov[..., N:, N:]  # [..., L, M, M]
        cov_cross = cov[..., :N, N:]  # [..., L, N, M]

        cov_shape = tf.shape(cov_add)
        noise = self.get_observation_noise() * tf.eye(
            cov_shape[-2], batch_shape=cov_shape[:-2], dtype=cov_add.dtype
        )
        L_add = tf.linalg.cholesky(cov_add + noise)  # [..., L, N, N]
        A = tf.linalg.triangular_solve(L_add, cov_cross, lower=True)  # [..., L, N, M]
        cov_qp_new = cov_qp - tf.matmul(A, A, transpose_a=True)  # [..., L, M, M]

        mean_add_diff = additional_data.observations - mean_add  # [..., N, L]
        mean_add_diff = leading_transpose(mean_add_diff, [..., -1, -2])[..., None]  # [..., L, N, 1]
        AM = tf.linalg.triangular_solve(L_add, mean_add_diff)  # [..., L, N, 1]
        mean_qp_new = mean_qp + leading_transpose(
            (tf.matmul(A, AM, transpose_a=True)[..., 0]), [..., -1, -2]
        )  # [..., M, L]

        tf.debugging.assert_shapes(
            [
                (additional_data.observations, [..., "N", "L"]),
                (query_points, ["M", "D"]),
                (mean_qp_new, [..., "M", "L"]),
                (cov_qp_new, [..., "L", "M", "M"]),
            ],
            message="received unexpected shapes computing conditional_predict_joint,"
            "check model kernel structure?",
        )

        return mean_qp_new, cov_qp_new

    def conditional_predict_f_sample_encoded(
        self, query_points: TensorType, additional_data: Dataset, num_samples: int
    ) -> TensorType:
        """
        Generates samples of the GP at query_points conditioned on both the model
        and some additional data.

        :param query_points: Set of query points with shape [M, D]
        :param additional_data: Dataset with query_points with shape [..., N, D] and observations
                 with shape [..., N, L]
        :param num_samples: number of samples
        :return: samples of f at query points, with shape [..., num_samples, M, L]
        """

        mean_new, cov_new = self.conditional_predict_joint(query_points, additional_data)
        mean_for_sample = tf.linalg.adjoint(mean_new)  # [..., L, N]
        samples = sample_mvn(
            mean_for_sample, cov_new, full_cov=True, num_samples=num_samples
        )  # [..., (S), P, N]
        return tf.linalg.adjoint(samples)  # [..., (S), N, L]

    def conditional_predict_y_encoded(
        self, query_points: TensorType, additional_data: Dataset
    ) -> tuple[TensorType, TensorType]:
        """
        Generates samples of y from the GP at query_points conditioned on both the model
        and some additional data.

        :param query_points: Set of query points with shape [M, D]
        :param additional_data: Dataset with query_points with shape [..., N, D] and observations
                 with shape [..., N, L]
        :return: predictive variance at query_points, with shape [..., M, L],
                 and predictive variance at query_points, with shape [..., M, L]
        """
        f_mean, f_var = self.conditional_predict_f(query_points, additional_data)
        return self.model.likelihood.predict_mean_and_var(query_points, f_mean, f_var)


class SparseGaussianProcessRegression(
    GPflowPredictor,
    EncodedSupportsCovarianceBetweenPoints,
    SupportsGetInducingVariables,
    SupportsGetInternalData,
    HasTrajectorySampler,
):
    """
    A :class:`TrainableProbabilisticModel` wrapper for a GPflow :class:`~gpflow.models.SGPR`.
    At the moment we only support models with a single latent GP. This is due to ``compute_qu``
    method in :class:`~gpflow.models.SGPR` that is used for computing covariance between
    query points and trajectory sampling, which at the moment works only for single latent GP.

    Similarly to our :class:`GaussianProcessRegression`, our :class:`~gpflow.models.SGPR` wrapper
    directly calls the posterior objects built by these models at prediction
    time. These posterior objects store the pre-computed Gram matrices, which can be reused to allow
    faster subsequent predictions. However, note that these posterior objects need to be updated
    whenever the underlying model is changed  by calling :meth:`update_posterior_cache` (this
    happens automatically after calls to :meth:`update` or :math:`optimize`).
    """

    def __init__(
        self,
        model: SGPR,
        optimizer: Optimizer | None = None,
        num_rff_features: int = 1000,
        inducing_point_selector: Optional[
            InducingPointSelector[SparseGaussianProcessRegression]
        ] = None,
        encoder: EncoderFunction | None = None,
    ):
        """
        :param model: The GPflow model to wrap.
        :param optimizer: The optimizer with which to train the model. Defaults to
            :class:`~trieste.models.optimizer.Optimizer` with :class:`~gpflow.optimizers.Scipy`.
        :param num_rff_features: The number of random Fourier features used to approximate the
            kernel when calling :meth:`trajectory_sampler`. We use a default of 1000 as it
            typically perfoms well for a wide range of kernels. Note that very smooth
            kernels (e.g. RBF) can be well-approximated with fewer features.
        :param inducing_point_selector: The (optional) desired inducing point selector that
            will update the underlying GPflow SGPR model's inducing points as
            the optimization progresses.
        :raise NotImplementedError (or ValueError): If we try to use a model with invalid
            ``num_rff_features``, or an ``inducing_point_selector`` with a model
            that has more than one set of inducing points.
        :param encoder: Optional encoder with which to transform query points before
            generating predictions.
        """
        super().__init__(optimizer, encoder)
        self._model = model

        check_optimizer(self.optimizer)

        if num_rff_features <= 0:
            raise ValueError(
                f"num_rff_features must be greater or equal to zero but got {num_rff_features}."
            )
        self._num_rff_features = num_rff_features

        if isinstance(self.model.inducing_variable, SeparateIndependentInducingVariables):
            if inducing_point_selector is not None:
                raise NotImplementedError(
                    f"""
                    InducingPointSelectors only currently support models with a single set
                    of inducing points however received inducing points of
                    type {type(self.model.inducing_variable)}.
                    """
                )
        self._inducing_point_selector = inducing_point_selector

        self._ensure_variable_model_data()
        self.create_posterior_cache()

    def __repr__(self) -> str:
        """"""
        return (
            f"SparseGaussianProcessRegression({self.model!r}, {self.optimizer!r},"
            f"{self._num_rff_features!r}, {self._inducing_point_selector!r})"
        )

    @property
    def model(self) -> SGPR:
        return self._model

    @property
    def inducing_point_selector(
        self,
    ) -> Optional[InducingPointSelector[SparseGaussianProcessRegression]]:
        return self._inducing_point_selector

    def predict_y_encoded(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
        f_mean, f_var = self.predict_encoded(query_points)
        return self.model.likelihood.predict_mean_and_var(query_points, f_mean, f_var)

    def _ensure_variable_model_data(self) -> None:
        # GPflow stores the data in Tensors. However, since we want to be able to update the data
        # without having to retrace the acquisition functions, put it in Variables instead.
        # Data has to be stored in variables with dynamic shape to allow for changes
        # Sometimes, for instance after serialization-deserialization, the shape can be overridden
        # Thus here we ensure data is stored in dynamic shape Variables
        if not all(is_variable(x) and x.shape[0] is None for x in self._model.data):
            self._model.data = (
                tf.Variable(
                    self._model.data[0],
                    trainable=False,
                    shape=[None, *self._model.data[0].shape[1:]],
                ),
                tf.Variable(
                    self._model.data[1],
                    trainable=False,
                    shape=[None, *self._model.data[1].shape[1:]],
                ),
            )

        if not is_variable(self._model.num_data):
            self._model.num_data = tf.Variable(self._model.num_data, trainable=False)

    def optimize_encoded(self, dataset: Dataset) -> OptimizeResult:
        """
        Optimize the model with the specified `dataset`.

        :param dataset: The data with which to optimize the `model`.
        """
        result = self.optimizer.optimize(self.model, dataset)
        self.update_posterior_cache()
        return result

    def update_encoded(self, dataset: Dataset) -> None:
        self._ensure_variable_model_data()

        x, y = self.model.data[0].value(), self.model.data[1].value()

        assert_data_is_compatible(dataset, Dataset(x, y))

        if dataset.query_points.shape[-1] != x.shape[-1]:
            raise ValueError

        if dataset.observations.shape[-1] != y.shape[-1]:
            raise ValueError

        self.model.data[0].assign(dataset.query_points)
        self.model.data[1].assign(dataset.observations)

        current_inducing_points, q_mu, _, _ = self.get_inducing_variables()

        if isinstance(current_inducing_points, list):
            inducing_points_trailing_dim = current_inducing_points[0].shape[-1]
        else:
            inducing_points_trailing_dim = current_inducing_points.shape[-1]

        if dataset.query_points.shape[-1] != inducing_points_trailing_dim:
            raise ValueError(
                f"Shape {dataset.query_points.shape} of new query points is incompatible with"
                f" shape {self.model.inducing_variable.Z.shape} of existing query points."
                f" Trailing dimensions must match."
            )

        if dataset.observations.shape[-1] != q_mu.shape[-1]:
            raise ValueError(
                f"Shape {dataset.observations.shape} of new observations is incompatible with"
                f" shape {self.model.q_mu.shape} of existing observations. Trailing"
                f" dimensions must match."
            )

        num_data = dataset.query_points.shape[0]
        self.model.num_data.assign(num_data)

        if self._inducing_point_selector is not None:
            new_inducing_points = self._inducing_point_selector.calculate_inducing_points(
                current_inducing_points, self, dataset
            )
            if not tf.reduce_all(
                tf.math.equal(
                    new_inducing_points,
                    current_inducing_points,
                )
            ):  # only bother updating if points actually change
                self._update_inducing_variables(new_inducing_points)

        self.update_posterior_cache()

    def _update_inducing_variables(self, new_inducing_points: TensorType) -> None:
        """
        When updating the inducing points of a model, we must also update the other
        inducing variables, i.e. `q_mu` and `q_sqrt` accordingly. The exact form of this update
        depends if we are using whitened representations of the inducing variables.
        See :meth:`_whiten_points` for details.

        :param new_inducing_points: The desired values for the new inducing points.
        :raise NotImplementedError: If we try to update the inducing variables of a model
            that has more than one set of inducing points.
        """

        if isinstance(new_inducing_points, list):
            raise NotImplementedError(
                f"""
                We do not currently support updating models with multiple sets of
                inducing points however received; {new_inducing_points}
                """
            )

        old_inducing_points, _, _, _ = self.get_inducing_variables()
        tf.assert_equal(
            tf.shape(old_inducing_points), tf.shape(new_inducing_points)
        )  # number of inducing points must not change

        if isinstance(self.model.inducing_variable, SharedIndependentInducingVariables):
            # gpflow says inducing_variable might be a ndarray; it won't
            cast(TensorType, self.model.inducing_variable.inducing_variable).Z.assign(
                new_inducing_points
            )  # [M, D]
        else:
            self.model.inducing_variable.Z.assign(new_inducing_points)  # [M, D]

    def get_inducing_variables(
        self,
    ) -> Tuple[Union[TensorType, list[TensorType]], TensorType, TensorType, bool]:
        """
        Return the model's inducing variables. The SGPR model does not have ``q_mu``, ``q_sqrt`` and
        ``whiten`` objects. We can use ``compute_qu`` method to obtain ``q_mu`` and ``q_sqrt``,
        while the SGPR model does not use the whitened representation. Note that at the moment
        ``compute_qu`` works only for single latent GP and returns ``q_sqrt`` in a shape that is
        inconsistent with the SVGP model (hence we need to do modify its shape).

        :return: The inducing points (i.e. locations of the inducing variables), as a Tensor or a
            list of Tensors (when the model has multiple inducing points); a tensor containing the
            variational mean ``q_mu``; a tensor containing the Cholesky decomposition of the
            variational covariance ``q_sqrt``; and a bool denoting if we are using whitened or
            non-whitened representations.
        :raise NotImplementedError: If the model has more than one latent GP.
        """
        if self.model.num_latent_gps > 1:
            raise NotImplementedError(
                f"""
                We do not currently support models with more than one latent GP,
                however received a model with {self.model.num_latent_gps} outputs.
                """
            )

        inducing_variable = self.model.inducing_variable

        if isinstance(inducing_variable, SharedIndependentInducingVariables):
            # gpflow says inducing_variable might be a ndarray; it won't
            inducing_points = cast(TensorType, inducing_variable.inducing_variable).Z  # [M, D]
        elif isinstance(inducing_variable, SeparateIndependentInducingVariables):
            inducing_points = [
                cast(TensorType, inducing_variable).Z
                for inducing_variable in inducing_variable.inducing_variables
            ]  # list of L [M, D] tensors
        else:
            inducing_points = inducing_variable.Z  # [M, D]

        q_mu, q_var = self.model.compute_qu()
        q_sqrt = tf.linalg.cholesky(q_var)
        q_sqrt = tf.expand_dims(q_sqrt, 0)
        whiten = False

        return inducing_points, q_mu, q_sqrt, whiten

    def covariance_between_points_encoded(
        self, query_points_1: TensorType, query_points_2: TensorType
    ) -> TensorType:
        r"""
        Compute the posterior covariance between sets of query points.

        Note that query_points_2 must be a rank 2 tensor, but query_points_1 can
        have leading dimensions.

        :param query_points_1: Set of query points with shape [..., A, D]
        :param query_points_2: Sets of query points with shape [B, D]
        :return: Covariance matrix between the sets of query points with shape [..., L, A, B]
            (L being the number of latent GPs = number of output dimensions)
        """
        inducing_points, _, q_sqrt, whiten = self.get_inducing_variables()

        return _covariance_between_points_for_variational_models(
            kernel=self.get_kernel(),
            inducing_points=inducing_points,
            q_sqrt=q_sqrt,
            query_points_1=query_points_1,
            query_points_2=query_points_2,
            whiten=whiten,
        )

    def trajectory_sampler(self) -> TrajectorySampler[SparseGaussianProcessRegression]:
        """
        Return a trajectory sampler. For :class:`SparseGaussianProcessRegression`, we build
        trajectories using a decoupled random Fourier feature approximation. Note that this
        is available only for single output models.

        At the moment only models with single latent GP are supported.

        :return: The trajectory sampler.
        :raise NotImplementedError: If we try to use the
            sampler with a model that has more than one latent GP.
        """
        if self.model.num_latent_gps > 1:
            raise NotImplementedError(
                f"""
                Trajectory sampler does not currently support models with multiple latent
                GPs, however received a model with {self.model.num_latent_gps} latent GPs.
                """
            )

        return DecoupledTrajectorySampler(self, self._num_rff_features)

    def get_internal_data(self) -> Dataset:
        """
        Return the model's training data.

        :return: The model's training data.
        """
        return Dataset(self.model.data[0], self.model.data[1])


class SparseVariational(
    GPflowPredictor,
    EncodedSupportsCovarianceBetweenPoints,
    SupportsGetInducingVariables,
    HasTrajectorySampler,
):
    """
    A :class:`TrainableProbabilisticModel` wrapper for a GPflow :class:`~gpflow.models.SVGP`.

    Similarly to our :class:`GaussianProcessRegression`, our :class:`~gpflow.models.SVGP` wrapper
    directly calls the posterior objects built by these models at prediction
    time. These posterior objects store the pre-computed Gram matrices, which can be reused to allow
    faster subsequent predictions. However, note that these posterior objects need to be updated
    whenever the underlying model is changed  by calling :meth:`update_posterior_cache` (this
    happens automatically after calls to :meth:`update` or :math:`optimize`).
    """

    def __init__(
        self,
        model: SVGP,
        optimizer: Optimizer | None = None,
        num_rff_features: int = 1000,
        inducing_point_selector: Optional[InducingPointSelector[SparseVariational]] = None,
        encoder: EncoderFunction | None = None,
    ):
        """
        :param model: The underlying GPflow sparse variational model.
        :param optimizer: The optimizer with which to train the model. Defaults to
            :class:`~trieste.models.optimizer.BatchOptimizer` with :class:`~tf.optimizers.Adam` with
            batch size 100.
        :param num_rff_features: The number of random Fourier features used to approximate the
            kernel when performing decoupled Thompson sampling through
            its :meth:`trajectory_sampler`. We use a default of 1000 as it typically
            perfoms well for a wide range of kernels. Note that very smooth kernels (e.g. RBF)
            can be well-approximated with fewer features.
        :param inducing_point_selector: The (optional) desired inducing_point_selector that
            will update the underlying GPflow sparse variational model's inducing points as
            the optimization progresses.
        :raise NotImplementedError: If we try to use an inducing_point_selector with a model
            that has more than one set of inducing points.
        :param encoder: Optional encoder with which to transform query points before
            generating predictions.
        """

        tf.debugging.assert_rank(
            model.q_sqrt, 3, "SparseVariational requires an SVGP model with q_diag=False."
        )

        if optimizer is None:
            optimizer = BatchOptimizer(tf_keras.optimizers.Adam(), batch_size=100, compile=True)

        super().__init__(optimizer, encoder)
        self._model = model

        if num_rff_features <= 0:
            raise ValueError(
                f"num_rff_features must be greater or equal to zero but got {num_rff_features}."
            )
        self._num_rff_features = num_rff_features

        check_optimizer(optimizer)

        if isinstance(self.model.inducing_variable, SeparateIndependentInducingVariables):
            if inducing_point_selector is not None:
                raise NotImplementedError(
                    f"""
                    InducingPointSelectors only currently support models with a single set
                    of inducing points however received inducing points of
                    type {type(self.model.inducing_variable)}.
                    """
                )

        self._inducing_point_selector = inducing_point_selector
        self._ensure_variable_model_data()
        self.create_posterior_cache()

    def _ensure_variable_model_data(self) -> None:
        # GPflow stores the data in Tensors. However, since we want to be able to update the data
        # without having to retrace the acquisition functions, put it in Variables instead.
        # Data has to be stored in variables with dynamic shape to allow for changes
        # Sometimes, for instance after serialization-deserialization, the shape can be overridden
        # Thus here we ensure data is stored in dynamic shape Variables
        if not is_variable(self._model.num_data):
            self._model.num_data = tf.Variable(self._model.num_data, trainable=False)

    def __repr__(self) -> str:
        """"""
        return (
            f"SparseVariational({self.model!r}, {self.optimizer!r},"
            f"{self._num_rff_features!r}, {self._inducing_point_selector!r})"
        )

    @property
    def model(self) -> SVGP:
        return self._model

    @property
    def inducing_point_selector(self) -> Optional[InducingPointSelector[SparseVariational]]:
        return self._inducing_point_selector

    def predict_y_encoded(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
        f_mean, f_var = self.predict_encoded(query_points)
        return self.model.likelihood.predict_mean_and_var(query_points, f_mean, f_var)

    def update_encoded(self, dataset: Dataset) -> None:
        self._ensure_variable_model_data()

        # Hard-code asserts from _assert_data_is_compatible because model doesn't store dataset
        current_inducing_points, q_mu, _, _ = self.get_inducing_variables()

        if isinstance(current_inducing_points, list):
            inducing_points_trailing_dim = current_inducing_points[0].shape[-1]
        else:
            inducing_points_trailing_dim = current_inducing_points.shape[-1]

        if dataset.query_points.shape[-1] != inducing_points_trailing_dim:
            raise ValueError(
                f"Shape {dataset.query_points.shape} of new query points is incompatible with"
                f" shape {self.model.inducing_variable.Z.shape} of existing query points."
                f" Trailing dimensions must match."
            )

        if dataset.observations.shape[-1] != q_mu.shape[-1]:
            raise ValueError(
                f"Shape {dataset.observations.shape} of new observations is incompatible with"
                f" shape {self.model.q_mu.shape} of existing observations. Trailing"
                f" dimensions must match."
            )

        num_data = dataset.query_points.shape[0]
        assert self.model.num_data is not None
        self.model.num_data.assign(num_data)

        if self._inducing_point_selector is not None:
            new_inducing_points = self._inducing_point_selector.calculate_inducing_points(
                current_inducing_points, self, dataset
            )
            if not tf.reduce_all(
                tf.math.equal(
                    new_inducing_points,
                    current_inducing_points,
                )
            ):  # only bother updating if points actually change
                self._update_inducing_variables(new_inducing_points)
                self.update_posterior_cache()

    def optimize_encoded(self, dataset: Dataset) -> OptimizeResult:
        """
        Optimize the model with the specified `dataset`.

        :param dataset: The data with which to optimize the `model`.
        """
        result = self.optimizer.optimize(self.model, dataset)
        self.update_posterior_cache()
        return result

    def _update_inducing_variables(self, new_inducing_points: TensorType) -> None:
        """
        When updating the inducing points of a model, we must also update the other
        inducing variables, i.e. `q_mu` and `q_sqrt` accordingly. The exact form of this update
        depends if we are using whitened representations of the inducing variables.
        See :meth:`_whiten_points` for details.

        :param new_inducing_points: The desired values for the new inducing points.
        :raise NotImplementedError: If we try to update the inducing variables of a model
            that has more than one set of inducing points.
        """

        if isinstance(new_inducing_points, list):
            raise NotImplementedError(
                f"""
                We do not currently support updating models with multiple sets of
                inducing points however received; {new_inducing_points}
                """
            )

        old_inducing_points, _, _, whiten = self.get_inducing_variables()
        tf.assert_equal(
            tf.shape(old_inducing_points), tf.shape(new_inducing_points)
        )  # number of inducing points must not change

        if whiten:
            new_q_mu, new_q_sqrt = _whiten_points(self, new_inducing_points)
        else:
            new_q_mu, new_f_cov = self.predict_joint_encoded(
                new_inducing_points
            )  # [N, L], [L, N, N]
            new_q_mu -= self.model.mean_function(new_inducing_points)
            jitter_mat = DEFAULTS.JITTER * tf.eye(
                tf.shape(new_inducing_points)[0], dtype=new_f_cov.dtype
            )
            new_q_sqrt = tf.linalg.cholesky(new_f_cov + jitter_mat)

        self.model.q_mu.assign(new_q_mu)  # [N, L]
        self.model.q_sqrt.assign(new_q_sqrt)  # [L, N, N]

        if isinstance(self.model.inducing_variable, SharedIndependentInducingVariables):
            # gpflow says inducing_variable might be a ndarray; it won't
            cast(TensorType, self.model.inducing_variable.inducing_variable).Z.assign(
                new_inducing_points
            )  # [M, D]
        else:
            self.model.inducing_variable.Z.assign(new_inducing_points)  # [M, D]

    def get_inducing_variables(
        self,
    ) -> Tuple[Union[TensorType, list[TensorType]], TensorType, TensorType, bool]:
        """
        Return the model's inducing variables.

        :return: The inducing points (i.e. locations of the inducing variables), as a Tensor or a
            list of Tensors (when the model has multiple inducing points); A tensor containing the
            variational mean q_mu; a tensor containing the Cholesky decomposition of the variational
            covariance q_sqrt; and a bool denoting if we are using whitened or
            non-whitened representations.
        """
        inducing_variable = self.model.inducing_variable

        if isinstance(inducing_variable, SharedIndependentInducingVariables):
            # gpflow says inducing_variable might be a ndarray; it won't
            inducing_points = cast(TensorType, inducing_variable.inducing_variable).Z  # [M, D]
        elif isinstance(inducing_variable, SeparateIndependentInducingVariables):
            inducing_points = [
                cast(TensorType, inducing_variable).Z
                for inducing_variable in inducing_variable.inducing_variables
            ]  # list of L [M, D] tensors
        else:
            inducing_points = inducing_variable.Z  # [M, D]

        return inducing_points, self.model.q_mu, self.model.q_sqrt, self.model.whiten

    def covariance_between_points_encoded(
        self, query_points_1: TensorType, query_points_2: TensorType
    ) -> TensorType:
        r"""
        Compute the posterior covariance between sets of query points.

        Note that query_points_2 must be a rank 2 tensor, but query_points_1 can
        have leading dimensions.

        :param query_points_1: Set of query points with shape [..., A, D]
        :param query_points_2: Sets of query points with shape [B, D]
        :return: Covariance matrix between the sets of query points with shape [..., L, A, B]
            (L being the number of latent GPs = number of output dimensions)
        """
        inducing_points, _, q_sqrt, whiten = self.get_inducing_variables()

        return _covariance_between_points_for_variational_models(
            kernel=self.get_kernel(),
            inducing_points=inducing_points,
            q_sqrt=q_sqrt,
            query_points_1=query_points_1,
            query_points_2=query_points_2,
            whiten=whiten,
        )

    def trajectory_sampler(self) -> TrajectorySampler[SparseVariational]:
        """
        Return a trajectory sampler. For :class:`SparseVariational`, we build
        trajectories using a decoupled random Fourier feature approximation.

        :return: The trajectory sampler.
        """
        return DecoupledTrajectorySampler(self, self._num_rff_features)


class VariationalGaussianProcess(
    GPflowPredictor,
    EncodedSupportsCovarianceBetweenPoints,
    SupportsGetInducingVariables,
    HasTrajectorySampler,
):
    r"""
    A :class:`TrainableProbabilisticModel` wrapper for a GPflow :class:`~gpflow.models.VGP`.

    A Variational Gaussian Process (VGP) approximates the posterior of a GP
    using the multivariate Gaussian closest to the posterior of the GP by minimizing the
    KL divergence between approximated and exact posteriors. See :cite:`opper2009variational`
    for details.

    The VGP provides (approximate) GP modelling under non-Gaussian likelihoods, for example
    when fitting a classification model over binary data.

    A whitened representation and (optional) natural gradient steps are used to aid
    model optimization.

    Similarly to our :class:`GaussianProcessRegression`, our :class:`~gpflow.models.VGP` wrapper
    directly calls the posterior objects built by these models at prediction
    time. These posterior objects store the pre-computed Gram matrices, which can be reused to allow
    faster subsequent predictions. However, note that these posterior objects need to be updated
    whenever the underlying model is changed  by calling :meth:`update_posterior_cache` (this
    happens automatically after calls to :meth:`update` or :math:`optimize`).
    """

    def __init__(
        self,
        model: VGP,
        optimizer: Optimizer | None = None,
        use_natgrads: bool = False,
        natgrad_gamma: Optional[float] = None,
        num_rff_features: int = 1000,
        encoder: EncoderFunction | None = None,
    ):
        """
        :param model: The GPflow :class:`~gpflow.models.VGP`.
        :param optimizer: The optimizer with which to train the model. Defaults to
            :class:`~trieste.models.optimizer.Optimizer` with :class:`~gpflow.optimizers.Scipy`.
        :param use_natgrads: If True then alternate model optimization steps with natural
            gradient updates. Note that natural gradients requires
            a :class:`~trieste.models.optimizer.BatchOptimizer` wrapper with
            :class:`~tf.optimizers.Optimizer` optimizer.
        :natgrad_gamma: Gamma parameter for the natural gradient optimizer.
        :param num_rff_features: The number of random Fourier features used to approximate the
            kernel when performing decoupled Thompson sampling through
            its :meth:`trajectory_sampler`. We use a default of 1000 as it typically perfoms
            well for a wide range of kernels. Note that very smooth kernels (e.g. RBF) can
            be well-approximated with fewer features.
        :raise ValueError (or InvalidArgumentError): If ``model``'s :attr:`q_sqrt` is not rank 3
            or if attempting to combine natural gradients with a :class:`~gpflow.optimizers.Scipy`
            optimizer.
        :param encoder: Optional encoder with which to transform query points before
            generating predictions.
        """
        tf.debugging.assert_rank(model.q_sqrt, 3)

        if optimizer is None and not use_natgrads:
            optimizer = Optimizer(gpflow.optimizers.Scipy(), compile=True)
        elif optimizer is None and use_natgrads:
            optimizer = BatchOptimizer(tf_keras.optimizers.Adam(), batch_size=100, compile=True)

        super().__init__(optimizer, encoder)

        check_optimizer(self.optimizer)

        if use_natgrads:
            if not isinstance(self.optimizer.optimizer, tf_keras.optimizers.Optimizer):
                raise ValueError(
                    f"""
                    Natgrads can only be used with a BatchOptimizer wrapper using an instance of
                    tf.optimizers.Optimizer, however received {self.optimizer}.
                    """
                )
            natgrad_gamma = 0.1 if natgrad_gamma is None else natgrad_gamma
        else:
            if isinstance(self.optimizer.optimizer, tf_keras.optimizers.Optimizer):
                raise ValueError(
                    f"""
                    If not using natgrads an Optimizer wrapper should be used with
                    gpflow.optimizers.Scipy, however received {self.optimizer}.
                    """
                )
            if natgrad_gamma is not None:
                raise ValueError(
                    """
                    natgrad_gamma is only to be specified when use_natgrads is True.
                    """
                )

        if num_rff_features <= 0:
            raise ValueError(
                f"num_rff_features must be greater or equal to zero but got {num_rff_features}."
            )
        self._num_rff_features = num_rff_features

        self._model = model
        self._use_natgrads = use_natgrads
        self._natgrad_gamma = natgrad_gamma
        self._ensure_variable_model_data()
        self.create_posterior_cache()

    def _ensure_variable_model_data(self) -> None:
        # GPflow stores the data in Tensors. However, since we want to be able to update the data
        # without having to retrace the acquisition functions, put it in Variables instead.
        # Data has to be stored in variables with dynamic shape to allow for changes
        # Sometimes, for instance after serialization-deserialization, the shape can be overridden
        # Thus here we ensure data is stored in dynamic shape Variables

        model = self.model
        if not all(isinstance(x, tf.Variable) and x.shape[0] is None for x in model.data):
            variable_data = (
                tf.Variable(
                    model.data[0],
                    trainable=False,
                    shape=[None, *model.data[0].shape[1:]],
                ),
                tf.Variable(
                    model.data[1],
                    trainable=False,
                    shape=[None, *model.data[1].shape[1:]],
                ),
            )

            # reinitialise the model so that the underlying Parameters have the right shape
            # and then reassign the original values
            old_q_mu = model.q_mu
            old_q_sqrt = model.q_sqrt
            model.__init__(  # type: ignore[misc]
                variable_data,
                model.kernel,
                model.likelihood,
                model.mean_function,
                model.num_latent_gps,
            )
            model.q_mu.assign(old_q_mu)
            model.q_sqrt.assign(old_q_sqrt)

    def __repr__(self) -> str:
        """"""
        return (
            f"VariationalGaussianProcess({self.model!r}, {self.optimizer!r})"
            f"{self._use_natgrads!r}, {self._natgrad_gamma!r}, {self._num_rff_features!r})"
        )

    @property
    def model(self) -> VGP:
        return self._model

    def predict_y_encoded(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
        f_mean, f_var = self.predict_encoded(query_points)
        return self.model.likelihood.predict_mean_and_var(query_points, f_mean, f_var)

    def update_encoded(self, dataset: Dataset, *, jitter: float = DEFAULTS.JITTER) -> None:
        """
        Update the model given the specified ``dataset``. Does not train the model.

        :param dataset: The data with which to update the model.
        :param jitter: The size of the jitter to use when stabilizing the Cholesky decomposition of
            the covariance matrix.
        """
        self._ensure_variable_model_data()
        update_vgp_data(self.model, (dataset.query_points, dataset.observations))
        self.update_posterior_cache()

    def optimize_encoded(self, dataset: Dataset) -> Optional[OptimizeResult]:
        """
        :class:`VariationalGaussianProcess` has a custom `optimize` method that (optionally) permits
        alternating between standard optimization steps (for kernel parameters) and natural gradient
        steps for the variational parameters (`q_mu` and `q_sqrt`). See :cite:`salimbeni2018natural`
        for details. Using natural gradients can dramatically speed up model fitting, especially for
        ill-conditioned posteriors.

        If using natural gradients, our optimizer inherits the mini-batch behavior and number
        of optimization steps as the base optimizer specified when initializing
        the :class:`VariationalGaussianProcess`.
        """
        model = self.model

        if self._use_natgrads:  # optimize variational params with natgrad optimizer
            natgrad_optimizer = gpflow.optimizers.NaturalGradient(gamma=self._natgrad_gamma)
            base_optimizer = self.optimizer

            gpflow.set_trainable(model.q_mu, False)  # variational params optimized by natgrad
            gpflow.set_trainable(model.q_sqrt, False)
            variational_params = [(model.q_mu, model.q_sqrt)]
            model_params = model.trainable_variables

            loss_fn = base_optimizer.create_loss(model, dataset)

            @jit(apply=self.optimizer.compile)
            def perform_optimization_step() -> None:  # alternate with natgrad optimizations
                natgrad_optimizer.minimize(loss_fn, variational_params)
                base_optimizer.optimizer.minimize(
                    loss_fn, model_params, **base_optimizer.minimize_args
                )

            for _ in range(base_optimizer.max_iter):  # type: ignore
                perform_optimization_step()

            gpflow.set_trainable(model.q_mu, True)  # revert variational params to trainable
            gpflow.set_trainable(model.q_sqrt, True)
            result = None  # TODO: find something useful to return
        else:
            result = self.optimizer.optimize(model, dataset)

        self.update_posterior_cache()
        return result

    def get_inducing_variables(self) -> Tuple[TensorType, TensorType, TensorType, bool]:
        """
        Return the model's inducing variables. Note that GPflow's VGP model is
        hard-coded to use the whitened representation.

        :return: Tensors containing: the inducing points (i.e. locations of the inducing
            variables); the variational mean q_mu; the Cholesky decomposition of the
            variational covariance q_sqrt; and a bool denoting if we are using whitened
            or non-whitened representations.
        """
        inducing_points = self.model.data[0]
        q_mu = self.model.q_mu
        q_sqrt = self.model.q_sqrt
        whiten = True  # GPflow's VGP model is hard-coded to use the whitened representation
        return inducing_points, q_mu, q_sqrt, whiten

    def trajectory_sampler(self) -> TrajectorySampler[VariationalGaussianProcess]:
        """
        Return a trajectory sampler. For :class:`VariationalGaussianProcess`, we build
        trajectories using a decoupled random Fourier feature approximation.

        At the moment only models with single latent GP are supported.

        :return: The trajectory sampler.
        :raise NotImplementedError: If we try to use the
            sampler with a model that has more than one latent GP.
        """
        if self.model.num_latent_gps > 1:
            raise NotImplementedError(
                f"""
                Trajectory sampler does not currently support models with multiple latent
                GPs, however received a model with {self.model.num_latent_gps} latent GPs.
                """
            )

        return DecoupledTrajectorySampler(self, self._num_rff_features)

    def covariance_between_points_encoded(
        self, query_points_1: TensorType, query_points_2: TensorType
    ) -> TensorType:
        r"""
        Compute the posterior covariance between sets of query points.

        Note that query_points_2 must be a rank 2 tensor, but query_points_1 can
        have leading dimensions.

        :param query_points_1: Set of query points with shape [..., A, D]
        :param query_points_2: Sets of query points with shape [B, D]
        :return: Covariance matrix between the sets of query points with shape [..., L, A, B]
            (L being the number of latent GPs = number of output dimensions)
        """

        _, _, q_sqrt, whiten = self.get_inducing_variables()

        return _covariance_between_points_for_variational_models(
            kernel=self.get_kernel(),
            inducing_points=self.model.data[0],
            q_sqrt=q_sqrt,
            query_points_1=query_points_1,
            query_points_2=query_points_2,
            whiten=whiten,
        )


class MultifidelityAutoregressive(
    TrainableProbabilisticModel, SupportsPredictY, SupportsCovarianceWithTopFidelity
):
    r"""
    A :class:`TrainableProbabilisticModel` implementation of the model
    from :cite:`Kennedy2000`. This is a multi-fidelity model that works with an
    arbitrary number of fidelities. It relies on there being a linear relationship
    between fidelities, and may not perform well for more complex relationships.
    Precisely, it models the relationship between sequential fidelities as

    .. math:: f_{i}(x) = \rho f_{i-1}(x) + \delta(x)

    where :math:`\rho` is a scalar and :math:`\delta` models the residual between the fidelities.
    The only base models supported in this implementation are :class:`~gpflow.models.GPR` models.
    Note: Currently only supports single output problems.
    """

    def __init__(
        self,
        fidelity_models: Sequence[GaussianProcessRegression],
    ):
        """
        :param fidelity_models: List of
            :class:`~trieste.models.gpflow.models.GaussianProcessRegression`
            models, one for each fidelity. The model at index 0 will be used as the signal model
            for the lowest fidelity and models at higher indices will be used as the residual
            model for each higher fidelity.
        """
        self._num_fidelities = len(fidelity_models)

        self.lowest_fidelity_signal_model = fidelity_models[0]
        # Note: The 0th index in the below is not a residual model, and should not be used.
        self.fidelity_residual_models: Sequence[GaussianProcessRegression] = fidelity_models
        # set this as a Parameter so that we can optimize it
        rho = [
            gpflow.Parameter(1.0, trainable=True, name=f"rho_{i}")
            for i in range(self.num_fidelities - 1)
        ]
        self.rho: list[gpflow.Parameter] = [
            gpflow.Parameter(1.0, trainable=False, name="dummy_variable"),
            *rho,
        ]

    @property
    def num_fidelities(self) -> int:
        return self._num_fidelities

    @inherit_check_shapes
    def predict(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
        """
        Predict the marginal mean and variance at query_points.

        :param query_points: Query points with shape [N, D+1], where the
            final column of the final dimension contains the fidelity of the query point
        :return: mean: The mean at query_points with shape [N, P],
                 and var: The variance at query_points with shape [N, P]
        """
        (
            query_points_wo_fidelity,  # [..., N, D]
            query_points_fidelity_col,  # [..., N, 1]
        ) = check_and_extract_fidelity_query_points(
            query_points, max_fidelity=self.num_fidelities - 1
        )

        signal_mean, signal_var = self.lowest_fidelity_signal_model.predict(
            query_points_wo_fidelity
        )  # [..., N, P], [..., N, P]

        for fidelity, (fidelity_residual_model, rho) in enumerate(
            zip(self.fidelity_residual_models, self.rho)
        ):
            if fidelity == 0:
                continue

            # Find indices of query points that need predicting for
            fidelity_float = tf.cast(fidelity, query_points.dtype)
            mask = query_points_fidelity_col >= fidelity_float  # [..., N, 1]
            fidelity_indices = tf.where(mask)[..., :-1]
            # Gather necessary query points and  predict
            fidelity_filtered_query_points = tf.gather_nd(
                query_points_wo_fidelity, fidelity_indices
            )
            (
                filtered_fidelity_residual_mean,
                filtered_fidelity_residual_var,
            ) = fidelity_residual_model.predict(fidelity_filtered_query_points)

            # Scatter predictions back into correct location
            fidelity_residual_mean = tf.tensor_scatter_nd_update(
                signal_mean, fidelity_indices, filtered_fidelity_residual_mean
            )
            fidelity_residual_var = tf.tensor_scatter_nd_update(
                signal_var, fidelity_indices, filtered_fidelity_residual_var
            )

            # Calculate mean and var for all columns (will be incorrect for qps with fid < fidelity)
            new_fidelity_signal_mean = rho * signal_mean + fidelity_residual_mean
            new_fidelity_signal_var = fidelity_residual_var + (rho**2) * signal_var

            # Mask out incorrect values and update mean and var for correct ones
            mask = query_points_fidelity_col >= fidelity_float
            signal_mean = tf.where(mask, new_fidelity_signal_mean, signal_mean)
            signal_var = tf.where(mask, new_fidelity_signal_var, signal_var)

        return signal_mean, signal_var

    def _calculate_residual(self, dataset: Dataset, fidelity: int) -> TensorType:
        r"""
        Calculate the true residuals for a set of datapoints at a given fidelity.

        Dataset should be made up of points that you have observations for at fidelity `fidelity`.
        The residuals calculated here are the difference between the data and the prediction at the
        lower fidelity multiplied by the rho value at this fidelity. This produces the training
        data for the residual models.

        .. math:: r_{i} = y - \rho_{i} * f_{i-1}(x)

        :param dataset: Dataset of points for which to calculate the residuals. Must have
         observations at fidelity `fidelity`. Query points shape is [N, D], observations is [N,P].
        :param fidelity: The fidelity for which to calculate the residuals
        :return: The true residuals at given datapoints for given fidelity, shape is [N,1].
        """
        fidelity_query_points = add_fidelity_column(dataset.query_points, fidelity - 1)
        residuals = (
            dataset.observations - self.rho[fidelity] * self.predict(fidelity_query_points)[0]
        )
        return residuals

    @inherit_check_shapes
    def sample(self, query_points: TensorType, num_samples: int) -> TensorType:
        """
        Sample `num_samples` samples from the posterior distribution at `query_points`

        :param query_points: The query points at which to sample of shape [N, D+1], where the
            final column of the final dimension contains the fidelity of the query point
        :param num_samples: The number of samples (S) to generate for each query point.
        :return: samples from the posterior of shape [..., S, N, P]
        """
        (
            query_points_wo_fidelity,
            query_points_fidelity_col,
        ) = check_and_extract_fidelity_query_points(
            query_points, max_fidelity=self.num_fidelities - 1
        )

        signal_sample = self.lowest_fidelity_signal_model.sample(
            query_points_wo_fidelity, num_samples
        )  # [S, N, P]

        for fidelity in range(1, int(tf.reduce_max(query_points_fidelity_col)) + 1):
            fidelity_residual_sample = self.fidelity_residual_models[fidelity].sample(
                query_points_wo_fidelity, num_samples
            )

            new_fidelity_signal_sample = (
                self.rho[fidelity] * signal_sample + fidelity_residual_sample
            )  # [S, N, P]

            mask = query_points_fidelity_col >= fidelity  # [N, P]
            mask = tf.broadcast_to(mask[..., None, :, :], new_fidelity_signal_sample.shape)

            signal_sample = tf.where(mask, new_fidelity_signal_sample, signal_sample)

        return signal_sample

    @inherit_check_shapes
    def predict_y(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
        """
        Predict the marginal mean and variance at `query_points` including observation noise

        :param query_points: Query points with shape [..., N, D+1], where the
            final column of the final dimension contains the fidelity of the query point
        :return: mean: The mean at query_points with shape [N, P],
                 and var: The variance at query_points with shape [N, P]
        """

        f_mean, f_var = self.predict(query_points)
        query_points_fidelity_col = query_points[..., -1:]

        # Get fidelity 0 observation noise
        observation_noise = (
            tf.ones_like(query_points_fidelity_col)
            * self.lowest_fidelity_signal_model.get_observation_noise()
        )

        for fidelity in range(1, self.num_fidelities):
            fidelity_observation_noise = (
                self.rho[fidelity] ** 2
            ) * observation_noise + self.fidelity_residual_models[fidelity].get_observation_noise()

            mask = query_points_fidelity_col >= fidelity

            observation_noise = tf.where(mask, fidelity_observation_noise, observation_noise)

        return f_mean, f_var + observation_noise

    def update(self, dataset: Dataset) -> None:
        """
        Update the models on their corresponding data. The data for each model is
        extracted by splitting the observations in ``dataset`` by fidelity level.

        :param dataset: The query points and observations for *all* the wrapped models.
        """
        check_and_extract_fidelity_query_points(
            dataset.query_points, max_fidelity=self.num_fidelities - 1
        )
        dataset_per_fidelity = split_dataset_by_fidelity(dataset, self.num_fidelities)
        for fidelity, dataset_for_fidelity in enumerate(dataset_per_fidelity):
            if fidelity == 0:
                self.lowest_fidelity_signal_model.update(dataset_for_fidelity)
            else:
                # Make query points but with final column corresponding to
                # fidelity we wish to predict at
                self.fidelity_residual_models[fidelity].update(
                    Dataset(
                        dataset_for_fidelity.query_points,
                        self._calculate_residual(dataset_for_fidelity, fidelity),
                    )
                )

    def optimize(self, dataset: Dataset) -> None:
        """
        Optimize all the models on their corresponding data. The data for each model is
        extracted by splitting the observations in ``dataset``  by fidelity level.
        Note that we have to code up a custom loss function when optimizing our residual
        model, so that we can include the correlation parameter as an optimisation variable.

        :param dataset: The query points and observations for *all* the wrapped models.
        """
        check_and_extract_fidelity_query_points(
            dataset.query_points, max_fidelity=self.num_fidelities - 1
        )
        dataset_per_fidelity = split_dataset_by_fidelity(dataset, self.num_fidelities)

        for fidelity, dataset_for_fidelity in enumerate(dataset_per_fidelity):
            if fidelity == 0:
                self.lowest_fidelity_signal_model.optimize(dataset_for_fidelity)
            else:
                gpf_residual_model = self.fidelity_residual_models[fidelity].model

                fidelity_observations = dataset_for_fidelity.observations
                fidelity_query_points = dataset_for_fidelity.query_points
                prev_fidelity_query_points = add_fidelity_column(
                    dataset_for_fidelity.query_points, fidelity - 1
                )
                predictions_from_lower_fidelity = self.predict(prev_fidelity_query_points)[0]

                def loss() -> TensorType:  # hardcoded log liklihood calculation for residual model
                    residuals = (
                        fidelity_observations - self.rho[fidelity] * predictions_from_lower_fidelity
                    )
                    K = gpf_residual_model.kernel(fidelity_query_points)
                    ks = add_noise_cov(K, gpf_residual_model.likelihood.variance)
                    L = tf.linalg.cholesky(ks)
                    m = gpf_residual_model.mean_function(fidelity_query_points)
                    log_prob = multivariate_normal(residuals, m, L)
                    return -1.0 * tf.reduce_sum(log_prob)

                trainable_variables = (
                    gpf_residual_model.trainable_variables + self.rho[fidelity].variables
                )
                self.fidelity_residual_models[fidelity].optimizer.optimizer.minimize(
                    loss, trainable_variables
                )
                residuals = self._calculate_residual(dataset_for_fidelity, fidelity)
                self.fidelity_residual_models[fidelity].update(
                    Dataset(fidelity_query_points, residuals)
                )

    def covariance_with_top_fidelity(self, query_points: TensorType) -> TensorType:
        """
        Calculate the covariance of the output at `query_point` and a given fidelity with the
        highest fidelity output at the same `query_point`.

        :param query_points: The query points to calculate the covariance for, of shape [N, D+1],
            where the final column of the final dimension contains the fidelity of the query point
        :return: The covariance with the top fidelity for the `query_points`, of shape [N, P]
        """
        fidelities = query_points[..., -1:]  # [..., 1]

        _, f_var = self.predict(query_points)

        for fidelity in range(self.num_fidelities - 1, -1, -1):
            mask = fidelities < fidelity

            f_var = tf.where(mask, f_var, f_var * self.rho[fidelity])

        return f_var

    def log(self, dataset: Optional[Dataset] = None) -> None:
        return


class MultifidelityNonlinearAutoregressive(
    TrainableProbabilisticModel, SupportsPredictY, SupportsCovarianceWithTopFidelity
):
    r"""
    A :class:`TrainableProbabilisticModel` implementation of the model from
    :cite:`perdikaris2017nonlinear`. This is a multifidelity model that works with
    an arbitrary number of fidelities. It is capable of modelling both linear and non-linear
    relationships between fidelities. It models the relationship between sequential fidelities as

    .. math:: f_{i}(x) =  g_{i}(x, f_{*i-1}(x))

    where :math:`f{*i-1}` is the posterior of the previous fidelity.
    The only base models supported in this implementation are :class:`~gpflow.models.GPR` models.
    Note: Currently only supports single output problems.
    """

    def __init__(
        self,
        fidelity_models: Sequence[GaussianProcessRegression],
        num_monte_carlo_samples: int = 100,
    ):
        """
        :param fidelity_models: List of
            :class:`~trieste.models.gpflow.models.GaussianProcessRegression`
            models, one for each fidelity. The model at index 0 should take
            inputs with the same number of dimensions as `x` and can use any kernel,
            whilst the later models should take an extra input dimesion, and use the kernel
            described in :cite:`perdikaris2017nonlinear`.
        :param num_monte_carlo_samples: The number of Monte Carlo samples to use for the
            sections of prediction and sampling that require the use of Monte Carlo methods.
        """

        self._num_fidelities = len(fidelity_models)
        self.fidelity_models = fidelity_models
        self.monte_carlo_random_numbers = tf.random.normal(
            [num_monte_carlo_samples, 1], dtype=tf.float64
        )

    @property
    def num_fidelities(self) -> int:
        return self._num_fidelities

    @inherit_check_shapes
    def sample(self, query_points: TensorType, num_samples: int) -> TensorType:
        """
        Return ``num_samples`` samples from the independent marginal distributions at
        ``query_points``.

        :param query_points: The points at which to sample, with shape [..., N, D].
        :param num_samples: The number of samples at each point.
        :return: The samples, with shape [..., S, N], where S is the number of samples.
        """
        (
            query_points_wo_fidelity,
            query_points_fidelity_col,
        ) = check_and_extract_fidelity_query_points(
            query_points, max_fidelity=self.num_fidelities - 1
        )  # [..., N, D], [..., N, 1]

        signal_sample = self.fidelity_models[0].sample(
            query_points_wo_fidelity, num_samples
        )  # [..., S, N, 1]

        # Repeat query_points to get same shape as signal sample
        query_points_fidelity_col = tf.broadcast_to(
            query_points_fidelity_col[..., None, :, :], signal_sample.shape
        )  # [..., S, N, 1]

        for fidelity in range(1, self.num_fidelities):
            qp_repeated = tf.broadcast_to(
                query_points_wo_fidelity[..., None, :, :],
                signal_sample.shape[:-1] + query_points_wo_fidelity.shape[-1],
            )  # [..., S, N, D]
            qp_augmented = tf.concat([qp_repeated, signal_sample], axis=-1)  # [..., S, N, D + 1]
            new_signal_sample = self.fidelity_models[fidelity].sample(
                qp_augmented, 1
            )  # [..., S, 1, N, 1]
            # Remove second dimension caused by getting a single sample
            new_signal_sample = new_signal_sample[..., :, 0, :, :]

            mask = query_points_fidelity_col >= fidelity

            signal_sample = tf.where(mask, new_signal_sample, signal_sample)

        return signal_sample

    @inherit_check_shapes
    def predict(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
        """
        Predict the marginal mean and variance at query_points.

        :param query_points: Query points with shape [..., N, D+1], where the
            final column of the final dimension contains the fidelity of the query point
        :return: mean: The mean at query_points with shape [..., N, P],
            and var: The variance at query_points with shape [..., N, P]
        """
        check_and_extract_fidelity_query_points(query_points, max_fidelity=self.num_fidelities - 1)

        sample_mean, sample_var = self._sample_mean_and_var_at_fidelities(
            query_points
        )  # [..., N, 1, S], [..., N, 1, S]
        variance = tf.reduce_mean(sample_var, axis=-1) + tf.math.reduce_variance(
            sample_mean, axis=-1
        )
        mean = tf.reduce_mean(sample_mean, axis=-1)
        return mean, variance

    @inherit_check_shapes
    def predict_y(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
        """
        Predict the marginal mean and variance at `query_points` including observation noise

        :param query_points: Query points with shape [..., N, D+1], where the
            final column of the final dimension contains the fidelity of the query point
        :return: mean: The mean at query_points with shape [N, P],
            and var: The variance at query_points with shape [N, P]
        """
        _, query_points_fidelity_col = check_and_extract_fidelity_query_points(
            query_points, max_fidelity=self.num_fidelities - 1
        )

        f_mean, f_var = self.predict(query_points)

        # Get fidelity 0 observation noise
        observation_noise = (
            tf.ones_like(query_points_fidelity_col)
            * self.fidelity_models[0].get_observation_noise()
        )

        for fidelity in range(1, self.num_fidelities):
            fidelity_observation_noise = self.fidelity_models[fidelity].get_observation_noise()

            mask = query_points_fidelity_col >= fidelity

            observation_noise = tf.where(mask, fidelity_observation_noise, observation_noise)
        return f_mean, f_var + observation_noise

    def _sample_mean_and_var_at_fidelities(
        self, query_points: TensorType
    ) -> tuple[TensorType, TensorType]:
        """
        Draw `num_monte_carlo_samples` samples of mean and variance from the model at the fidelities
        passed in the final column of the query points.

        :param query_points:  Query points with shape [..., N, D+1], where the
            final column of the final dimension contains the fidelity of the query point
        :return: sample_mean: Samples of the mean at the query points with shape [..., N, 1, S]
            and sample_var: Samples of the variance at the query points with shape [..., N, 1, S]
        """

        (
            query_points_wo_fidelity,
            query_points_fidelity_col,
        ) = check_and_extract_fidelity_query_points(
            query_points, max_fidelity=self.num_fidelities - 1
        )  # [..., N, D], [..., N, 1]

        sample_mean, sample_var = self.fidelity_models[0].predict(
            query_points_wo_fidelity
        )  # [..., N, 1], [..., N, 1]

        # Create new dimension to store samples for each query point
        # Repeat the inital sample mean and variance S times and add a dimension in the
        # middle (so that sample mean and query points can be concatenated sensibly)
        sample_mean = tf.broadcast_to(
            sample_mean, sample_mean.shape[:-1] + self.monte_carlo_random_numbers.shape[0]
        )[
            ..., :, None, :
        ]  # [..., N, 1, S]
        sample_var = tf.broadcast_to(
            sample_var, sample_var.shape[:-1] + self.monte_carlo_random_numbers.shape[0]
        )[
            ..., :, None, :
        ]  # [..., N, 1, S]

        # Repeat fidelity points for each sample to match shapes for masking
        query_points_fidelity_col = tf.broadcast_to(
            query_points_fidelity_col,
            query_points_fidelity_col.shape[:-1] + self.monte_carlo_random_numbers.shape[0],
        )[
            ..., :, None, :
        ]  # [..., N, 1, S]

        # Predict for all fidelities but stop updating once we have
        # reached desired fidelity for each query point
        for fidelity in range(1, self.num_fidelities):
            # sample_mean [..., N, 1, S]
            # sample_var [..., N, 1, S]
            (
                next_fidelity_sample_mean,
                next_fidelity_sample_var,
            ) = self._propagate_samples_through_level(
                query_points_wo_fidelity, fidelity, sample_mean, sample_var
            )

            mask = query_points_fidelity_col >= fidelity  # [..., N, 1, S]

            sample_mean = tf.where(mask, next_fidelity_sample_mean, sample_mean)  # [..., N, 1, S]
            sample_var = tf.where(mask, next_fidelity_sample_var, sample_var)  # [..., N, 1, S]

        return sample_mean, sample_var

    def _propagate_samples_through_level(
        self,
        query_point: TensorType,
        fidelity: int,
        sample_mean: TensorType,
        sample_var: TensorType,
    ) -> tuple[TensorType, TensorType]:
        """
        Propagate samples through a given fidelity.

        This takes a set of query points without a fidelity column and calculates samples
        at the given fidelity, using the sample means and variances from the previous fidelity.

        :param query_points: The query points to sample at, with no fidelity column,
            with shape[..., N, D]
        :param fidelity: The fidelity to propagate the samples through
        :param sample_mean: Samples of the posterior mean at the previous fidelity,
            with shape [..., N, 1, S]
        :param sample_var: Samples of the posterior variance at the previous fidelity,
            with shape [..., N, 1, S]
        :return: sample_mean: Samples of the posterior mean at the given fidelity,
            of shape [..., N, 1, S]
            and sample_var: Samples of the posterior variance at the given fidelity,
            of shape [..., N, 1, S]
        """
        # Repeat random numbers for each query point and add middle dimension
        # for concatentation with query points This also means that it has the same
        # shape as sample_var and sample_mean, so there's no broadcasting required
        # Note: at the moment we use the same monte carlo values for every value in the batch dim

        reshaped_random_numbers = tf.broadcast_to(
            tf.transpose(self.monte_carlo_random_numbers)[..., None, :],
            sample_mean.shape,
        )  # [..., N, 1, S]
        samples = reshaped_random_numbers * tf.sqrt(sample_var) + sample_mean  # [..., N, 1, S]
        # Add an extra unit dim to query_point and repeat for each of the samples
        qp_repeated = tf.broadcast_to(
            query_point[..., :, :, None],  # [..., N, D, 1]
            query_point.shape + samples.shape[-1],
        )  # [..., N, D, S]
        qp_augmented = tf.concat([qp_repeated, samples], axis=-2)  # [..., N, D+1, S]

        # Flatten sample dimension to n_qp dimension to pass through predictor
        # Switch dims to make reshape match up correct dimensions for query points
        # Use Einsum to switch last two dimensions
        qp_augmented = tf.linalg.matrix_transpose(qp_augmented)  # [..., N, S, D+1]

        flat_qp_augmented, unflatten = flatten_leading_dims(qp_augmented)  # [...*N*S, D+1]

        # Dim of flat qp augmented is now [n_qps*n_samples, qp_dims], as the model expects
        sample_mean, sample_var = self.fidelity_models[fidelity].predict(
            flat_qp_augmented
        )  # [...*N*S, 1], [...*N*S, 1]

        # Reshape back to have samples as final dimension
        sample_mean = unflatten(sample_mean)  # [..., N, S, 1]
        sample_var = unflatten(sample_var)  # [..., N, S, 1]
        sample_mean = tf.linalg.matrix_transpose(sample_mean)  # [..., N, 1, S]
        sample_var = tf.linalg.matrix_transpose(sample_var)  # [..., N, 1, S]

        return sample_mean, sample_var

    def update(self, dataset: Dataset) -> None:
        """
        Update the models on their corresponding data. The data for each model is
        extracted by splitting the observations in ``dataset`` by fidelity level.

        :param dataset: The query points and observations for *all* the wrapped models.
        """
        check_and_extract_fidelity_query_points(
            dataset.query_points, max_fidelity=self.num_fidelities - 1
        )
        dataset_per_fidelity = split_dataset_by_fidelity(
            dataset, num_fidelities=self.num_fidelities
        )
        for fidelity, dataset_for_fidelity in enumerate(dataset_per_fidelity):
            if fidelity == 0:
                self.fidelity_models[0].update(dataset_for_fidelity)
            else:
                cur_fidelity_model = self.fidelity_models[fidelity]
                new_final_query_point_col, _ = self.predict(
                    add_fidelity_column(dataset_for_fidelity.query_points, fidelity - 1)
                )
                new_query_points = tf.concat(
                    [dataset_for_fidelity.query_points, new_final_query_point_col], axis=1
                )
                cur_fidelity_model.update(
                    Dataset(new_query_points, dataset_for_fidelity.observations)
                )

    def optimize(self, dataset: Dataset) -> None:
        """
        Optimize all the models on their corresponding data. The data for each model is
        extracted by splitting the observations in ``dataset``  by fidelity level.

        :param dataset: The query points and observations for *all* the wrapped models.
        """
        check_and_extract_fidelity_query_points(
            dataset.query_points, max_fidelity=self.num_fidelities - 1
        )
        dataset_per_fidelity = split_dataset_by_fidelity(dataset, self.num_fidelities)

        for fidelity, dataset_for_fidelity in enumerate(dataset_per_fidelity):
            if fidelity == 0:
                self.fidelity_models[0].optimize(dataset_for_fidelity)
            else:
                fidelity_observations = dataset_for_fidelity.observations
                fidelity_query_points = dataset_for_fidelity.query_points
                prev_fidelity_query_points = add_fidelity_column(
                    fidelity_query_points, fidelity - 1
                )
                means_from_lower_fidelity = self.predict(prev_fidelity_query_points)[0]
                augmented_qps = tf.concat(
                    [fidelity_query_points, means_from_lower_fidelity], axis=1
                )
                self.fidelity_models[fidelity].optimize(
                    Dataset(augmented_qps, fidelity_observations)
                )
                self.fidelity_models[fidelity].update(Dataset(augmented_qps, fidelity_observations))

    def covariance_with_top_fidelity(self, query_points: TensorType) -> TensorType:
        """
        Calculate the covariance of the output at `query_point` and a given fidelity with the
        highest fidelity output at the same `query_point`.

        :param query_points: The query points to calculate the covariance for, of shape [N, D+1],
            where the final column of the final dimension contains the fidelity of the query point
        :return: The covariance with the top fidelity for the `query_points`, of shape [N, P]
        """
        num_samples = 100
        (
            query_points_wo_fidelity,
            query_points_fidelity_col,
        ) = check_and_extract_fidelity_query_points(
            query_points, max_fidelity=self.num_fidelities - 1
        )  # [N, D], [N, 1]

        # Signal sample stops updating once fidelity is reached for that query point
        signal_sample = self.fidelity_models[0].model.predict_f_samples(
            query_points_wo_fidelity, num_samples, full_cov=False
        )

        # Repeat query_points to get same shape as signal sample
        query_points_fidelity_col = tf.broadcast_to(
            query_points_fidelity_col[None, :, :], signal_sample.shape
        )
        # Max fidelity sample keeps updating to the max fidelity
        max_fidelity_sample = tf.identity(signal_sample)

        for fidelity in range(1, self.num_fidelities):
            qp_repeated = tf.broadcast_to(
                query_points_wo_fidelity[None, :, :],
                tf.TensorShape(num_samples) + query_points_wo_fidelity.shape,
            )  # [S, N, D]
            # We use max fidelity sample here, which is okay because anything
            # with a lower fidelity will not be updated
            qp_augmented = tf.concat([qp_repeated, max_fidelity_sample], axis=-1)  # [S, N, D + 1]
            new_signal_sample = self.fidelity_models[fidelity].model.predict_f_samples(
                qp_augmented, 1, full_cov=False
            )
            # Remove second dimension caused by getting a single sample
            new_signal_sample = new_signal_sample[:, 0, :, :]

            mask = query_points_fidelity_col >= fidelity

            signal_sample = tf.where(mask, new_signal_sample, signal_sample)  # [S, N, 1]
            max_fidelity_sample = new_signal_sample  # [S, N , 1]

        cov = tfp.stats.covariance(signal_sample, max_fidelity_sample)[:, :, 0]

        return cov

    def log(self, dataset: Optional[Dataset] = None) -> None:
        return



# Copyright 2020 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" This module contains utilities for :class:`~trieste.observer.Observer` data. """
from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, Sequence

import tensorflow as tf

from trieste.types import TensorType


@dataclass(frozen=True)
class Dataset:
    """
    Container for the query points and corresponding observations from an
    :class:`~trieste.observer.Observer`.
    """

    query_points: TensorType
    """ The points at which the :class:`~trieste.observer.Observer` was queried. """

    observations: TensorType
    """ The observed output of the :class:`~trieste.observer.Observer` for each query point. """

    def __post_init__(self) -> None:
        """
        :raise ValueError (or InvalidArgumentError): If ``query_points`` or ``observations`` have \
            rank less than two, or they have unequal shape in any but their last dimension.
        """
        tf.debugging.assert_rank_at_least(self.query_points, 2)
        tf.debugging.assert_rank_at_least(self.observations, 2)

        if 0 in (self.query_points.shape[-1], self.observations.shape[-1]):
            raise ValueError(
                f"query_points and observations cannot have dimension 0, got shapes"
                f" {self.query_points.shape} and {self.observations.shape}."
            )

        if (
            self.query_points.shape[:-1] != self.observations.shape[:-1]
            # can't check dynamic shapes, so trust that they're ok (if not, they'll fail later)
            and None not in self.query_points.shape[:-1]
            and None not in self.observations.shape[:-1]
        ):
            raise ValueError(
                f"Leading shapes of query_points and observations must match. Got shapes"
                f" {self.query_points.shape}, {self.observations.shape}."
            )

    def __add__(self, rhs: Dataset) -> Dataset:
        r"""
        Return the :class:`Dataset` whose query points are the result of concatenating the
        `query_points` in each :class:`Dataset` along the zeroth axis, and the same for the
        `observations`. For example:

        >>> d1 = Dataset(
        ...     tf.constant([[0.1, 0.2], [0.3, 0.4]]),
        ...     tf.constant([[0.5, 0.6], [0.7, 0.8]])
        ... )
        >>> d2 = Dataset(tf.constant([[0.9, 1.0]]), tf.constant([[1.1, 1.2]]))
        >>> (d1 + d2).query_points
        <tf.Tensor: shape=(3, 2), dtype=float32, numpy=
        array([[0.1, 0.2],
               [0.3, 0.4],
               [0.9, 1. ]], dtype=float32)>
        >>> (d1 + d2).observations
        <tf.Tensor: shape=(3, 2), dtype=float32, numpy=
        array([[0.5, 0.6],
               [0.7, 0.8],
               [1.1, 1.2]], dtype=float32)>

        :param rhs: A :class:`Dataset` with the same shapes as this one, except in the zeroth
            dimension, which can have any size.
        :return: The result of concatenating the :class:`Dataset`\ s.
        :raise InvalidArgumentError: If the shapes of the `query_points` in each :class:`Dataset`
            differ in any but the zeroth dimension. The same applies for `observations`.
        """
        return Dataset(
            tf.concat([self.query_points, rhs.query_points], axis=0),
            tf.concat([self.observations, rhs.observations], axis=0),
        )

    def __len__(self) -> tf.Tensor:
        """
        :return: The number of query points, or equivalently the number of observations.
        """
        return tf.shape(self.observations)[0]

    def __deepcopy__(self, memo: dict[int, object]) -> Dataset:
        return self

    def astuple(self) -> tuple[TensorType, TensorType]:
        """
        **Note:** Unlike the standard library function `dataclasses.astuple`, this method does
        **not** deepcopy the attributes.

        :return: A 2-tuple of the :attr:`query_points` and :attr:`observations`.
        """
        return self.query_points, self.observations


def check_and_extract_fidelity_query_points(
    query_points: TensorType, max_fidelity: Optional[int] = None
) -> tuple[TensorType, TensorType]:
    """Check whether the final column of a tensor is close enough to ints
    to be reasonably considered to represent fidelities.

    The final input column of multi-fidelity data should be a reference to
    the fidelity of the query point. We cannot have mixed type tensors, but
    we can check that thhe final column values are suitably close to integers.

    :param query_points: Data to check final column of.
    :raise: ValueError: If there are not enough columns to be multifidelity data
    :raise InvalidArgumentError: If any value in the final column is far from an integer
    :return: Query points without fidelity column
        and the fidelities of each of the query points
    """
    # Check we have sufficient columns
    if query_points.shape[-1] < 2:
        raise ValueError(
            "Query points do not have enough columns to be multifidelity,"
            f" need at least 2, got {query_points.shape[1]}"
        )
    input_points = query_points[..., :-1]
    fidelity_col = query_points[..., -1:]
    # Check fidelity column values are close to ints
    tf.debugging.assert_equal(
        tf.round(fidelity_col),
        fidelity_col,
        message="Fidelity column should be float(int), but got a float that"
        " was not close to an int",
    )
    # Check fidelity column values are non-negative

    tf.debugging.assert_non_negative(fidelity_col, message="Fidelity must be non-negative")
    if max_fidelity is not None:
        max_input_fid = tf.reduce_max(fidelity_col)
        max_fidelity_float = tf.cast(max_fidelity, dtype=query_points.dtype)
        tf.debugging.assert_less_equal(
            max_input_fid,
            max_fidelity_float,
            message=(
                f"Model only supports fidelities up to {max_fidelity},"
                f" but {max_input_fid} was passed"
            ),
        )

    return input_points, fidelity_col


def split_dataset_by_fidelity(dataset: Dataset, num_fidelities: int) -> Sequence[Dataset]:
    """Split dataset into individual datasets without fidelity information

    :param dataset: Dataset for which to split fidelities
    :param num_fidlities: Number of fidelities in the problem (not just dataset)
    :return: Ordered list of datasets with lowest fidelity at index 0 and highest at -1
    """
    if num_fidelities < 1:
        raise ValueError(f"Data must have 1 or more fidelities, got {num_fidelities}")
    datasets = [get_dataset_for_fidelity(dataset, fidelity) for fidelity in range(num_fidelities)]
    return datasets


def get_dataset_for_fidelity(dataset: Dataset, fidelity: int) -> Dataset:
    """Get a dataset with only the specified fidelity of data in

    :param dataset: The dataset from which to extract the single fidelity data
    :param fidelity: The fidelity to extract the data for
    :return: Dataset with a single fidelity and no fidelity column
    """

    input_points, fidelity_col = check_and_extract_fidelity_query_points(
        dataset.query_points
    )  # [..., D], [..., 1]
    mask = fidelity_col == fidelity  # [..., ]
    inds = tf.where(mask)[..., 0]  # [..., ]
    inputs_for_fidelity = tf.gather(input_points, inds, axis=0)  # [..., D]
    observations_for_fidelity = tf.gather(dataset.observations, inds, axis=0)  # [..., 1]
    return Dataset(query_points=inputs_for_fidelity, observations=observations_for_fidelity)


def add_fidelity_column(query_points: TensorType, fidelity: int) -> TensorType:
    """Add fidelity column to query_points without fidelity data

    :param query_points: query points without fidelity to add fidelity column to
    :param fidelity: fidelity to populate fidelity column with
    :return: TensorType of query points with fidelity column added
    """
    fidelity_col = tf.ones((tf.shape(query_points)[-2], 1), dtype=query_points.dtype) * fidelity
    query_points_for_fidelity = tf.concat([query_points, fidelity_col], axis=-1)
    return query_points_for_fidelity


def split_dataset_randomly(
    dataset: Dataset, proportion: float, seed: Optional[int] = None
) -> tuple[Dataset, Dataset]:
    """Split dataset randomly into two parts according to the specified proportion.

    :param dataset: Dataset to split.
    :param proportion: Proportion of data to include in first dataset (between 0 and 1).
    :param seed: Optional random seed for reproducibility.
    :return: Tuple of two datasets, where first contains approximately proportion * len(dataset)
        points.
    :raise ValueError: If proportion is not between 0 and 1
    """
    if not 0 <= proportion <= 1:
        raise ValueError(f"proportion must be between 0 and 1, got {proportion}")

    if len(dataset) == 0:
        return dataset, dataset

    if seed is not None:  # ensure reproducibility
        tf.random.set_seed(seed)

    # Generate random permutation of indices
    num_points = len(dataset)
    indices = tf.random.shuffle(tf.range(num_points), seed=seed)

    # Split indices according to proportion
    split_idx = tf.cast(tf.math.floor(tf.cast(num_points, tf.float32) * proportion), tf.int32)
    first_indices = indices[:split_idx]
    second_indices = indices[split_idx:]

    # Create new datasets using the split indices
    first_dataset = Dataset(
        tf.gather(dataset.query_points, first_indices),
        tf.gather(dataset.observations, first_indices),
    )
    second_dataset = Dataset(
        tf.gather(dataset.query_points, second_indices),
        tf.gather(dataset.observations, second_indices),
    )

    return first_dataset, second_dataset



# Copyright 2020 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from abc import abstractmethod
from typing import Optional, Type, cast

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.figure import Figure
from typing_extensions import Protocol

from ...space import SearchSpace
from ...types import TensorType
from .plotting import create_grid


class Simulation(Protocol):
    """A representation of a constrained objective used for plotting."""

    threshold: float

    @staticmethod
    @abstractmethod
    def objective(input_data: TensorType) -> TensorType:
        """Objective function."""

    @staticmethod
    @abstractmethod
    def constraint(input_data: TensorType) -> TensorType:
        """Constraint function."""


def plot_objective_and_constraints(
    search_space: SearchSpace, simulation: Type[Simulation]
) -> Figure:
    """
    Plot constrained objective.

    :param search_space: Search space
    :param simulation: Constrained objective
    :return: figure
    """
    objective_fn = simulation.objective
    constraint_fn = simulation.constraint
    lower_bound = search_space.lower
    upper_bound = search_space.upper

    grid, xx, yy = create_grid(lower_bound, upper_bound, grid_density=30)
    objective = objective_fn(grid).numpy()
    constraint = constraint_fn(grid).numpy()
    fig, (axes1, axes2) = plt.subplots(2, 2, sharex="all", sharey="all", figsize=(8, 8))

    levels = 30

    axes1[0].contourf(xx, yy, objective.reshape(*xx.shape), levels, alpha=0.9)
    axes1[1].contourf(xx, yy, constraint.reshape(*xx.shape), levels, alpha=0.9)
    axes1[0].set_title("Objective")
    axes1[1].set_title("Constraint")

    mask_ids = np.argwhere(constraint > simulation.threshold)
    mask = np.zeros_like(objective, dtype=bool)
    mask[mask_ids] = True
    objective_masked = np.ma.array(objective, mask=mask)
    constraint_masked = np.ma.array(constraint, mask=mask)

    axes2[0].contourf(xx, yy, objective_masked.reshape(*xx.shape), levels, alpha=0.9)
    axes2[1].contourf(xx, yy, constraint_masked.reshape(*xx.shape), levels, alpha=0.9)
    axes2[0].set_title("Constrained objective")
    axes2[1].set_title("Constraint mask")

    for ax in np.ravel([axes1, axes2]):
        ax.set_xlim(lower_bound[0], upper_bound[0])
        ax.set_ylim(lower_bound[1], upper_bound[1])

    return fig


def plot_init_query_points(
    search_space: SearchSpace,
    simulation: Type[Simulation],
    objective_data: TensorType,
    constraint_data: TensorType,
    new_constraint_data: Optional[TensorType] = None,
) -> Figure:
    """
    Plot initial query points on constrained objective.

    :param search_space: Search space
    :param simulation: Constrained objective
    :param objective_data: Objective data
    :param constraint_data: Constraint data
    :param new_constraint_data: Optional new constraint data
    :return: figure
    """
    objective_fn = simulation.objective
    constraint_fn = simulation.constraint
    lower_bound = search_space.lower
    upper_bound = search_space.upper
    levels = 30
    psize = 15
    cw, cb, co = "white", "tab:blue", "tab:orange"

    grid, xx, yy = create_grid(lower_bound, upper_bound, grid_density=30)
    objective = objective_fn(grid).numpy()
    constraint = constraint_fn(grid).numpy()
    fig, ax = plt.subplots(1, 1, figsize=(8, 6))
    constraint_points = constraint_data[-1].numpy()
    x = objective_data[0].numpy()

    mask_ids = constraint[:, 0] > simulation.threshold
    mask = np.zeros_like(objective, dtype=bool)
    mask[mask_ids, :] = True
    objective_masked = np.ma.array(objective, mask=mask)

    def in_out_points(
        x: TensorType, constraint_points: TensorType
    ) -> tuple[TensorType, TensorType]:
        ids_in = constraint_points[:, 0] <= simulation.threshold
        ids_out = constraint_points[:, 0] > simulation.threshold
        return x.T[..., ids_in], x.T[..., ids_out]

    (x_in, y_in), (x_out, y_out) = in_out_points(x, constraint_points)

    ax.contourf(xx, yy, objective_masked.reshape(*xx.shape), levels, alpha=0.9)
    ax.scatter(x_in, y_in, s=psize, c=cb, edgecolors=cw, marker="o")
    ax.scatter(x_out, y_out, s=psize, c=cw, edgecolors=cb, marker="o")

    if new_constraint_data is not None:
        x_new, constraint_points_new = new_constraint_data
        (x_in_new, y_in_new), (x_out_new, y_out_new) = in_out_points(
            x_new.numpy(), constraint_points_new.numpy()
        )
        ax.scatter(x_in_new, y_in_new, s=psize, c=co, edgecolors=cw, marker="o")
        ax.scatter(x_out_new, y_out_new, s=psize, c=cw, edgecolors=co, marker="o")

    ax.set_title("Constrained objective")
    ax.set_xlim(lower_bound[0], upper_bound[0])
    ax.set_ylim(lower_bound[1], upper_bound[1])

    return fig


def plot_2obj_cst_query_points(
    search_space: SearchSpace,
    simulation: Type[Simulation],
    objective_data: TensorType,
    constraint_data: TensorType,
) -> None:
    """
    Plot 2 objective constrainted query points.

    :param search_space: Search space
    :param simulation: Constrained objective
    :param objective_data: Objective data
    :param constraint_data: Constraint data
    """

    class Sim1(simulation):  # type: ignore
        @staticmethod
        def objective(input_data: TensorType) -> TensorType:
            return simulation.objective(input_data)[:, 0:1]

    class Sim2(simulation):  # type: ignore
        @staticmethod
        def objective(input_data: TensorType) -> TensorType:
            return simulation.objective(input_data)[:, 1:2]

    for sim in [Sim1, Sim2]:
        plot_init_query_points(
            search_space,
            cast(Type[Simulation], sim),
            objective_data,
            constraint_data,
        )



# Copyright 2021 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import re
from typing import Any, Dict, Mapping, Optional

import dill
import tensorflow as tf
import tensorflow_probability as tfp
import tensorflow_probability.python.distributions as tfd
from check_shapes import check_shapes
from gpflow.keras import tf_keras
from tensorflow.python.keras.callbacks import Callback

from ... import logging
from ...data import Dataset, split_dataset_randomly
from ...space import EncoderFunction
from ...types import TensorType
from ...utils import flatten_leading_dims
from ..interfaces import (
    EncodedSupportsPredictY,
    EncodedTrainableProbabilisticModel,
    HasTrajectorySampler,
    TrajectorySampler,
)
from ..optimizer import KerasOptimizer
from ..utils import write_summary_data_based_metrics
from .architectures import KerasEnsemble, MultivariateNormalTriL
from .interface import DeepEnsembleModel, KerasPredictor
from .sampler import DeepEnsembleTrajectorySampler
from .utils import negative_log_likelihood, sample_model_index, sample_with_replacement


class DeepEnsemble(
    KerasPredictor,
    EncodedTrainableProbabilisticModel,
    DeepEnsembleModel,
    HasTrajectorySampler,
    EncodedSupportsPredictY,
):
    """
    A :class:`~trieste.model.TrainableProbabilisticModel` wrapper for deep ensembles built using
    Keras.

    Deep ensembles are ensembles of deep neural networks that have been found to have good
    representation of uncertainty in practice (<cite data-cite="lakshminarayanan2017simple"/>).
    This makes them a potentially attractive model for Bayesian optimization for use-cases with
    large number of observations, non-stationary objective functions and need for fast predictions,
    in which standard Gaussian process models are likely to struggle. The model consists of simple
    fully connected multilayer probabilistic networks as base learners, with Gaussian distribution
    as a final layer, using the negative log-likelihood loss for training the networks. The
    model relies on differences in random initialization of weights for generating diversity among
    base learners.

    The original formulation of the model does not include boostrapping of the data. The authors
    found that it does not improve performance the model. We include bootstrapping as an option
    as later work that more precisely measured uncertainty quantification found that boostrapping
    does help with uncertainty representation (see <cite data-cite="osband2021epistemic"/>).

    We provide classes for constructing ensembles using Keras
    (:class:`~trieste.models.keras.KerasEnsemble`) in the `architectures` package that should be
    used with the :class:`~trieste.models.keras.DeepEnsemble` wrapper. There we also provide a
    :class:`~trieste.models.keras.GaussianNetwork` base learner following the original
    formulation in <cite data-cite="lakshminarayanan2017simple"/>, but any user-specified network
    can be supplied, as long as it has a Gaussian distribution as a final layer and follows the
    :class:`~trieste.models.keras.KerasEnsembleNetwork` interface.

    A word of caution in case a learning rate scheduler is used in ``fit_args`` to
    :class:`KerasOptimizer` optimizer instance. Typically one would not want to continue with the
    reduced learning rate in the subsequent Bayesian optimization step. Hence, we reset the
    learning rate to the original one after calling the ``fit`` method. In case this is not the
    behaviour you would like, you will need to subclass the model and overwrite the
    :meth:`optimize` method.

    Currently, we do not support setting up the model with dictionary config.
    """

    def __init__(
        self,
        model: KerasEnsemble,
        optimizer: Optional[KerasOptimizer] = None,
        bootstrap: bool = False,
        diversify: bool = False,
        continuous_optimisation: bool = True,
        compile_args: Optional[Mapping[str, Any]] = None,
        encoder: EncoderFunction | None = None,
    ) -> None:
        """
        :param model: A Keras ensemble model with probabilistic networks as ensemble members. The
            model has to be built but not compiled.
        :param optimizer: The optimizer wrapper with necessary specifications for compiling and
            training the model. Defaults to :class:`~trieste.models.optimizer.KerasOptimizer` with
            :class:`~tf.optimizers.Adam` optimizer, negative log likelihood loss, mean squared
            error metric and a dictionary of default arguments for Keras `fit` method: 3000 epochs,
            batch size 16, early stopping callback with patience of 50, and verbose 0.
            See https://keras.io/api/models/model_training_apis/#fit-method for a list of possible
            arguments.
        :param bootstrap: Sample with replacement data for training each network in the ensemble.
            By default, set to `False`.
        :param diversify: Whether to use quantiles from the approximate Gaussian distribution of
            the ensemble as trajectories instead of mean predictions when calling
            :meth:`trajectory_sampler`. This mode can be used to increase the diversity
            in case of optimizing very large batches of trajectories. By
            default, set to `False`.
        :param continuous_optimisation: If True (default), the optimizer will keep track of the
            number of epochs across BO iterations and use this number as initial_epoch. This is
            essential to allow monitoring of model training across BO iterations.
        :param compile_args: Keyword arguments to pass to the ``compile`` method of the
            Keras model (:class:`~tf.keras.Model`).
            See https://keras.io/api/models/model_training_apis/#compile-method for a
            list of possible arguments. The ``optimizer``, ``loss`` and ``metrics`` arguments
            must not be included.
        :param encoder: Optional encoder with which to transform query points before
            generating predictions.
        :raise ValueError: If ``model`` is not an instance of
            :class:`~trieste.models.keras.KerasEnsemble`, or ensemble has less than two base
            learners (networks), or `compile_args` contains disallowed arguments.
        """
        if model.ensemble_size < 2:
            raise ValueError(f"Ensemble size must be greater than 1 but got {model.ensemble_size}.")

        super().__init__(optimizer, encoder)

        if compile_args is None:
            compile_args = {}

        if not {"optimizer", "loss", "metrics"}.isdisjoint(compile_args):
            raise ValueError(
                "optimizer, loss and metrics arguments must not be included in compile_args."
            )

        if not self.optimizer.fit_args:
            self.optimizer.fit_args = {
                "verbose": 0,
                "epochs": 3000,
                "batch_size": 16,
                "callbacks": [
                    tf_keras.callbacks.EarlyStopping(
                        monitor="loss", patience=50, restore_best_weights=True
                    )
                ],
            }

        if self.optimizer.loss is None:
            self.optimizer.loss = negative_log_likelihood

        if self.optimizer.metrics is None:
            self.optimizer.metrics = ["mse"]

        model.model.compile(
            optimizer=self.optimizer.optimizer,
            loss=[self.optimizer.loss] * model.ensemble_size,
            metrics=[self.optimizer.metrics] * model.ensemble_size,
            **compile_args,
        )

        if not isinstance(
            self.optimizer.optimizer.lr, tf_keras.optimizers.schedules.LearningRateSchedule
        ):
            self.original_lr = self.optimizer.optimizer.lr.numpy()
        self._absolute_epochs = 0
        self._continuous_optimisation = continuous_optimisation

        self._model = model
        self._bootstrap = bootstrap
        self._diversify = diversify

    def __repr__(self) -> str:
        """"""
        return (
            f"DeepEnsemble({self.model!r}, {self.optimizer!r}, {self._bootstrap!r}, "
            f"{self._diversify!r})"
        )

    @property
    def model(self) -> tf_keras.Model:
        """Returns compiled Keras ensemble model."""
        return self._model.model

    @property
    def ensemble_size(self) -> int:
        """
        Returns the size of the ensemble, that is, the number of base learners or individual neural
        network models in the ensemble.
        """
        return self._model.ensemble_size

    @property
    def num_outputs(self) -> int:
        """
        Returns the number of outputs trained on by each member network.
        """
        return self._model.num_outputs

    def prepare_dataset(
        self, dataset: Dataset, do_not_bootstrap: bool = False
    ) -> tuple[Dict[str, TensorType], Dict[str, TensorType]]:
        """
        Transform ``dataset`` into inputs and outputs with correct names that can be used for
        training the :class:`KerasEnsemble` model.

        If ``bootstrap`` argument in the :class:`~trieste.models.keras.DeepEnsemble` is set to
        `True`, data will be additionally sampled with replacement, independently for
        each network in the ensemble.

        :param dataset: A dataset with ``query_points`` and ``observations`` tensors.
        :param do_not_bootstrap: If True, do not bootstrap the data, even if the model is set to
            bootstrap. This can be useful for preparing validation data.
        :return: A dictionary with input data and a dictionary with output data.
        """
        inputs = {}
        outputs = {}
        for index in range(self.ensemble_size):
            if self._bootstrap and not do_not_bootstrap:
                resampled_data = sample_with_replacement(dataset)
            else:
                resampled_data = dataset
            input_name = self.model.input_names[index]
            output_name = self.model.output_names[index]
            inputs[input_name], outputs[output_name] = resampled_data.astuple()

        return inputs, outputs

    def prepare_query_points(self, query_points: TensorType) -> Dict[str, TensorType]:
        """
        Transform ``query_points`` into inputs with correct names that can be used for
        predicting with the model.

        :param query_points: A tensor with ``query_points``.
        :return: A dictionary with query_points prepared for predictions.
        """
        inputs = {}
        for index in range(self.ensemble_size):
            inputs[self.model.input_names[index]] = query_points

        return inputs

    def ensemble_distributions(self, query_points: TensorType) -> tuple[tfd.Distribution, ...]:
        """
        Return distributions for each member of the ensemble.

        :param query_points: The points at which to return distributions.
        :return: The distributions for the observations at the specified
            ``query_points`` for each member of the ensemble.
        """
        x_transformed: dict[str, TensorType] = self.prepare_query_points(query_points)
        return self._model.model(x_transformed)

    def predict_encoded(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
        r"""
        Returns mean and variance at ``query_points`` for the whole ensemble. Variance here is
        the epistemic uncertainty, computed as the variance of the ensemble means.

        Following <cite data-cite="lakshminarayanan2017simple"/> we treat the ensemble as a
        uniformly-weighted Gaussian mixture model and combine the predictions as

        .. math:: p(y|\mathbf{x}) = M^{-1} \Sum_{m=1}^M \mathcal{N}
            (\mu_{\theta_m}(\mathbf{x}),\,\sigma_{\theta_m}^{2}(\mathbf{x}))

        We further approximate the ensemble prediction as a Gaussian whose mean and variance
        are respectively the mean and variance of the mixture, given by

        .. math:: \mu_{*}(\mathbf{x}) = M^{-1} \Sum_{m=1}^M \mu_{\theta_m}(\mathbf{x})

        .. math:: \sigma^2_{*}(\mathbf{x}) = M^{-1} \Sum_{m=1}^M (\mu^2_{\theta_m}(\mathbf{x}))
            - \mu^2_{*}(\mathbf{x})

        This method assumes that the final layer in each member of the ensemble is
        probabilistic, an instance of :class:`~tfp.distributions.Distribution`. In particular, given
        the nature of the approximations stated above the final layer should be a Gaussian
        distribution with `mean` and `variance` methods.

        :param query_points: The points at which to make predictions.
        :return: The predicted mean and variance of the observations at the specified
            ``query_points``.
        """
        ensemble_means, _ = self.predict_ensemble_encoded(query_points)
        predicted_means = tf.math.reduce_mean(ensemble_means, axis=-3)
        epistemic_variance = tf.math.reduce_variance(ensemble_means, axis=-3)

        return predicted_means, epistemic_variance

    def predict_y_encoded(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
        r"""
        Returns mean and variance at ``query_points`` for the whole ensemble. Variance here is
        the total uncertainty, computed as the sum of the epistemic and aleatoric uncertainties.

        Following <cite data-cite="lakshminarayanan2017simple"/> we treat the ensemble as a
        uniformly-weighted Gaussian mixture model and combine the predictions as

        .. math:: p(y|\mathbf{x}) = M^{-1} \Sum_{m=1}^M \mathcal{N}
            (\mu_{\theta_m}(\mathbf{x}),\,\sigma_{\theta_m}^{2}(\mathbf{x}))

        We further approximate the ensemble prediction as a Gaussian whose mean and variance
        are respectively the mean and variance of the mixture, given by

        .. math:: \mu_{*}(\mathbf{x}) = M^{-1} \Sum_{m=1}^M \mu_{\theta_m}(\mathbf{x})

        .. math:: \sigma^2_{*}(\mathbf{x}) = M^{-1} \Sum_{m=1}^M (\sigma_{\theta_m}^{2}(\mathbf{x})
            + \mu^2_{\theta_m}(\mathbf{x})) - \mu^2_{*}(\mathbf{x})

        This method assumes that the final layer in each member of the ensemble is
        probabilistic, an instance of :class:`~tfp.distributions.Distribution`. In particular, given
        the nature of the approximations stated above the final layer should be a Gaussian
        distribution with `mean` and `variance` methods.

        :param query_points: The points at which to make predictions.
        :return: The predicted mean and variance of the observations at the specified
            ``query_points``.
        """
        ensemble_means, ensemble_vars = self.predict_ensemble_encoded(query_points)
        predicted_means = tf.math.reduce_mean(ensemble_means, axis=-3)
        epistemic_variance = tf.math.reduce_variance(ensemble_means, axis=-3)
        aleatoric_variance = tf.math.reduce_mean(ensemble_vars, axis=-3)

        return predicted_means, epistemic_variance + aleatoric_variance

    def predict_noise_encoded(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
        """
        Returns mean and variance of the noise at ``query_points`` for the whole ensemble.
        Mean is the mean of the variance across the ensemble, variance is the variance of the
        variance across the ensemble. Assumes that the data is encoded already.

        :param query_points: The points at which to make predictions.
        :return: The predicted mean and variance of the aleatoric noise at the specified
            ``query_points``.
        """
        _, ensemble_vars = self.predict_ensemble_encoded(query_points)
        aleatoric_variance_mean = tf.math.reduce_mean(ensemble_vars, axis=-3)
        aleatoric_variance_variance = tf.math.reduce_variance(ensemble_vars, axis=-3)

        return aleatoric_variance_mean, aleatoric_variance_variance

    @check_shapes(
        "query_points: [batch..., D]",
        "return[0]: [batch..., E...]",
        "return[1]: [batch..., E...]",
    )
    def predict_noise(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
        """
        Returns mean and variance of the noise at ``query_points`` for the whole ensemble.
        Mean is the mean of the variance across the ensemble, variance is the variance of the
        variance across the ensemble. Data is encoded before prediction if encoder is provided.

        :param query_points: The points at which to make predictions.
        :return: The predicted mean and variance of the aleatoric noise at the specified
            ``query_points``.
        """
        return self.predict_noise_encoded(self.encode(query_points))

    @property
    def dtype(self) -> tf.DType:
        """The prediction dtype."""
        return self._model.output_dtype

    def predict_ensemble_encoded(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
        """
        Returns mean and variance at ``query_points`` for each member of the ensemble. First tensor
        is the mean and second is the variance, where each has shape [..., M, N, 1], where M is
        the ``ensemble_size``.

        This method assumes that the final layer in each member of the ensemble is
        probabilistic, an instance of :class:`¬tfp.distributions.Distribution`, in particular
        `mean` and `variance` methods should be available.

        :param query_points: The points at which to make predictions.
        :return: The predicted mean and variance of the observations at the specified
            ``query_points`` for each member of the ensemble.
        """
        input_dims = min(len(query_points.shape), len(self.model.input_shape[0]))
        flat_x, unflatten = flatten_leading_dims(query_points, output_dims=input_dims)
        ensemble_distributions = self.ensemble_distributions(flat_x)
        predicted_means = tf.stack(
            [unflatten(dist.mean()) for dist in ensemble_distributions], axis=-3
        )
        predicted_vars = tf.stack(
            [unflatten(dist.variance()) for dist in ensemble_distributions], axis=-3
        )

        return predicted_means, predicted_vars

    def predict_ensemble(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
        return self.predict_ensemble_encoded(self.encode(query_points))

    def sample_encoded(self, query_points: TensorType, num_samples: int) -> TensorType:
        """
        Return ``num_samples`` samples at ``query_points``. We use the mixture approximation in
        :meth:`predict` for ``query_points`` and sample ``num_samples`` times from a Gaussian
        distribution given by the predicted mean and variance.

        :param query_points: The points at which to sample, with shape [..., N, D].
        :param num_samples: The number of samples at each point.
        :return: The samples. For a predictive distribution with event shape E, this has shape
            [..., S, N] + E, where S is the number of samples.
        """

        predicted_means, predicted_vars = self.predict_encoded(query_points)
        normal = tfp.distributions.Normal(predicted_means, tf.sqrt(predicted_vars))
        samples = normal.sample(num_samples)

        return samples  # [num_samples, len(query_points), 1]

    def sample_ensemble_encoded(self, query_points: TensorType, num_samples: int) -> TensorType:
        """
        Return ``num_samples`` samples at ``query_points``. Each sample is taken from a Gaussian
        distribution given by the predicted mean and variance of a randomly chosen network in the
        ensemble. This avoids using the Gaussian mixture approximation and samples directly from
        individual Gaussian distributions given by each network in the ensemble.

        :param query_points: The points at which to sample, with shape [..., N, D].
        :param num_samples: The number of samples at each point.
        :return: The samples. For a predictive distribution with event shape E, this has shape
            [..., S, N] + E, where S is the number of samples.
        """
        ensemble_distributions = self.ensemble_distributions(query_points)
        network_indices = sample_model_index(self.ensemble_size, num_samples)

        stacked_samples = []
        for i in range(num_samples):
            stacked_samples.append(ensemble_distributions[network_indices[i]].sample())
        samples = tf.stack(stacked_samples, axis=0)

        return samples  # [num_samples, len(query_points), 1]

    def sample_ensemble(self, query_points: TensorType, num_samples: int) -> TensorType:
        return self.sample_ensemble_encoded(self.encode(query_points), num_samples)

    def trajectory_sampler(self) -> TrajectorySampler[DeepEnsemble]:
        """
        Return a trajectory sampler. For :class:`DeepEnsemble`, we use an ensemble
        sampler that randomly picks a network from the ensemble and uses its predicted means
        for generating a trajectory, or optionally randomly sampled quantiles rather than means.

        :return: The trajectory sampler.
        """
        return DeepEnsembleTrajectorySampler(self, self._diversify)

    def update_encoded(self, dataset: Dataset) -> None:
        """
        Neural networks are parametric models and do not need to update data.
        `TrainableProbabilisticModel` interface, however, requires an update method, so
        here we simply pass the execution.
        """
        return

    def optimize_encoded(self, dataset: Dataset) -> tf_keras.callbacks.History:
        """
        Optimize the underlying Keras ensemble model with the specified ``dataset``.

        Optimization is performed by using the Keras `fit` method, rather than applying the
        optimizer and using the batches supplied with the optimizer wrapper. User can pass
        arguments to the `fit` method through ``minimize_args`` argument in the optimizer wrapper.
        These default to using 100 epochs, batch size 100, and verbose 0. See
        https://keras.io/api/models/model_training_apis/#fit-method for a list of possible
        arguments.

        Note that optimization does not return the result, instead optimization results are
        stored in a history attribute of the model object.

        :param dataset: The data with which to optimize the model.
        """
        fit_args = dict(self.optimizer.fit_args)

        # Tell optimizer how many epochs have been used before: the optimizer will "continue"
        # optimization across multiple BO iterations rather than start fresh at each iteration.
        # This allows us to monitor training across iterations.

        if "epochs" in fit_args:
            fit_args["epochs"] = fit_args["epochs"] + self._absolute_epochs

        if "validation_split" in fit_args:
            validation_dataset, dataset = split_dataset_randomly(
                dataset, fit_args["validation_split"]
            )
            x_val, y_val = self.prepare_dataset(validation_dataset, do_not_bootstrap=True)
            del fit_args["validation_split"]
            fit_args["validation_data"] = tf.data.Dataset.from_tensor_slices((x_val, y_val))

        x, y = self.prepare_dataset(dataset)
        tf_train_dataset = self._build_tf_dataset(x, y)

        history = self.model.fit(
            tf_train_dataset,
            **fit_args,
            initial_epoch=self._absolute_epochs,
        )
        if self._continuous_optimisation:
            self._absolute_epochs = self._absolute_epochs + len(history.history["loss"])

        # Reset lr in case there was an lr schedule: a schedule will have changed the learning
        # rate, so that the next time we call `optimize` the starting learning rate would be
        # different. Therefore, we make sure the learning rate is set back to its initial value.
        # However, this is not needed for `LearningRateSchedule` instances.
        if not isinstance(
            self.optimizer.optimizer.lr, tf_keras.optimizers.schedules.LearningRateSchedule
        ):
            self.optimizer.optimizer.lr.assign(self.original_lr)

        return history

    def _build_tf_dataset(
        self, x: Dict[str, tf.Tensor], y: Dict[str, tf.Tensor]
    ) -> tf.data.Dataset:
        tf_dataset = tf.data.Dataset.from_tensor_slices((x, y))

        if "steps_per_epoch" in self.optimizer.fit_args:
            tf_dataset = tf_dataset.prefetch(tf.data.experimental.AUTOTUNE).repeat()

        if "batch_size" in self.optimizer.fit_args:
            batch_size = self.optimizer.fit_args["batch_size"]
            tf_dataset = tf_dataset.batch(batch_size)

        return tf_dataset

    def log(self, dataset: Optional[Dataset] = None) -> None:
        """
        Log model training information at a given optimization step to the Tensorboard.
        We log several summary statistics of losses and metrics given in ``fit_args`` to
        ``optimizer`` (final, difference between inital and final loss, min and max). We also log
        epoch statistics, but as histograms, rather than time series. We also log several training
        data based metrics, such as root mean square error between predictions and observations,
        and several others.

        We do not log statistics of individual models in the ensemble unless specifically switched
        on with ``trieste.logging.set_summary_filter(lambda name: True)``.

        For custom logs user will need to subclass the model and overwrite this method.

        :param dataset: Optional data that can be used to log additional data-based model summaries.
        """
        summary_writer = logging.get_tensorboard_writer()
        if summary_writer:
            with summary_writer.as_default(step=logging.get_step_number()):
                logging.scalar("epochs/num_epochs", len(self.model.history.epoch))
                for k, v in self.model.history.history.items():
                    KEY_SPLITTER = {
                        # map history keys to prefix and suffix
                        "loss": ("loss", ""),
                        r"(?P<model>model_\d+)_output_loss": ("loss", r"_\g<model>"),
                        r"(?P<model>model_\d+)_output_(?P<metric>.+)": (
                            r"\g<metric>",
                            r"_\g<model>",
                        ),
                    }
                    for pattern, (pre_sub, post_sub) in KEY_SPLITTER.items():
                        if re.match(pattern, k):
                            pre = re.sub(pattern, pre_sub, k)
                            post = re.sub(pattern, post_sub, k)
                            break
                    else:
                        # unrecognised history key; ignore
                        continue
                    if "model" in post:
                        if not logging.include_summary("_ensemble"):
                            break
                        pre = pre + "/_ensemble"
                    logging.histogram(f"{pre}/epoch{post}", lambda: v)
                    logging.scalar(f"{pre}/final{post}", lambda: v[-1])
                    logging.scalar(f"{pre}/diff{post}", lambda: v[0] - v[-1])
                    logging.scalar(f"{pre}/min{post}", lambda: tf.reduce_min(v))
                    logging.scalar(f"{pre}/max{post}", lambda: tf.reduce_max(v))
                if dataset:
                    write_summary_data_based_metrics(
                        dataset=dataset, model=self, prefix="training_"
                    )
                    if logging.include_summary("_ensemble"):
                        predict_ensemble_variance = self.predict_ensemble(dataset.query_points)[1]
                        for i in range(predict_ensemble_variance.shape[0]):
                            logging.histogram(
                                f"variance/_ensemble/predict_variance_model_{i}",
                                predict_ensemble_variance[i, ...],
                            )
                            logging.scalar(
                                f"variance/_ensemble/predict_variance_mean_model_{i}",
                                tf.reduce_mean(predict_ensemble_variance[i, ...]),
                            )

    def __getstate__(self) -> dict[str, Any]:
        # use to_json and get_weights to save any optimizer fit_arg callback models
        state = self.__dict__.copy()
        if self._optimizer:
            callbacks: list[Callback] = self._optimizer.fit_args.get("callbacks", [])
            saved_models: list[KerasOptimizer] = []
            tensorboard_writers: list[dict[str, Any]] = []
            try:
                for callback in callbacks:
                    # serialize the callback models before pickling the optimizer
                    saved_models.append(callback.model)
                    if callback.model is self.model:
                        # no need to serialize the main model, just use a special value instead
                        callback.model = ...
                    elif callback.model:
                        callback.model = (callback.model.to_json(), callback.model.get_weights())
                    # don't pickle tensorboard writers either; they'll be recreated when needed
                    if isinstance(callback, tf_keras.callbacks.TensorBoard):
                        tensorboard_writers.append(callback._writers)
                        callback._writers = {}
                state["_optimizer"] = dill.dumps(state["_optimizer"])
            except Exception as e:
                raise NotImplementedError(
                    "Failed to copy DeepEnsemble optimizer due to unsupported callbacks."
                ) from e
            finally:
                # revert original state, even if the pickling failed
                for callback, model in zip(callbacks, saved_models):
                    callback.model = model
                for callback, writers in zip(
                    (cb for cb in callbacks if isinstance(cb, tf_keras.callbacks.TensorBoard)),
                    tensorboard_writers,
                ):
                    callback._writers = writers

        # don't serialize any history optimization result
        if isinstance(state.get("_last_optimization_result"), tf_keras.callbacks.History):
            state["_last_optimization_result"] = ...

        return state

    def __setstate__(self, state: dict[str, Any]) -> None:
        # Restore optimizer and callback models after depickling, and recompile.
        self.__dict__.update(state)

        # Unpickle the optimizer, and restore all the callback models
        self._optimizer = dill.loads(self._optimizer)
        for callback in self._optimizer.fit_args.get("callbacks", []):
            if callback.model is ...:
                callback.set_model(self.model)
            elif callback.model:
                model_json, weights = callback.model
                model = tf_keras.models.model_from_json(
                    model_json,
                    custom_objects={"MultivariateNormalTriL": MultivariateNormalTriL},
                )
                model.set_weights(weights)
                callback.set_model(model)

        # Recompile the model
        self.model.compile(
            self.optimizer.optimizer,
            loss=[self.optimizer.loss] * self._model.ensemble_size,
            metrics=[self.optimizer.metrics] * self._model.ensemble_size,
        )

        # recover optimization result if necessary (and possible)
        if state.get("_last_optimization_result") is ...:
            self._last_optimization_result = getattr(self.model, "history")



# Copyright 2020 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module is the home of  Trieste's functionality for choosing the inducing points
of sparse variational Gaussian processes (i.e. our :class:`SparseVariational` wrapper).
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generic

import gpflow
import tensorflow as tf
import tensorflow_probability as tfp
from scipy.cluster.vq import kmeans

from ...data import Dataset
from ...space import Box, DiscreteSearchSpace, SearchSpace
from ...types import TensorType
from ..interfaces import ProbabilisticModelType
from .interface import GPflowPredictor


class InducingPointSelector(ABC, Generic[ProbabilisticModelType]):
    """
    This class provides functionality to update the inducing points of an inducing point-based model
    as the Bayesian optimization progresses.

    The only constraint on subclasses of :class:`InducingPointSelector` is that they preserve
    the shape of the inducing points so not to trigger expensive retracing.

    It can often be beneficial to change the inducing points during optimization, for example
    to allow the model to focus its limited modelling resources into promising areas of the space.
    See :cite:`vakili2021scalable` for demonstrations of some of
    our :class:`InducingPointSelectors`.
    """

    def __init__(self, recalc_every_model_update: bool = True):
        """
        :param recalc_every_model_update: If True then recalculate the inducing points for each
            model update, otherwise just recalculate on the first call.
        """
        self._recalc_every_model_update = recalc_every_model_update
        self._initialized = False

    def calculate_inducing_points(
        self,
        current_inducing_points: TensorType,
        model: ProbabilisticModelType,
        dataset: Dataset,
    ) -> TensorType:
        """
        Calculate the new inducing points given the existing inducing points.

        If `recalc_every_model_update` is set to False then we only generate new inducing points
        for the first :meth:`calculate_inducing_points` call, otherwise we just return the current
        inducing points.

        :param current_inducing_points: The current inducing points used by the model.
        :param model: The sparse model.
        :param dataset: The data from the observer.
        :return: The new updated inducing points.
        :raise NotImplementedError: If model has more than one set of inducing variables.
        """
        tf.debugging.Assert(current_inducing_points is not None, [tf.constant([])])

        if isinstance(current_inducing_points, list):
            raise NotImplementedError(
                """
                InducingPointSelectors only currently support models with a single set
                of inducing points.
                """
            )

        if (
            not self._initialized
        ) or self._recalc_every_model_update:  # calculate new inducing points when required
            self._initialized = True
            M = tf.shape(current_inducing_points)[0]
            new_inducing_points = self._recalculate_inducing_points(M, model, dataset)  # [M, D]
            tf.assert_equal(tf.shape(current_inducing_points), tf.shape(new_inducing_points))
            return new_inducing_points  # [M, D]
        else:  # otherwise dont recalculate
            return current_inducing_points  # [M, D]

    @abstractmethod
    def _recalculate_inducing_points(
        self, M: int, model: ProbabilisticModelType, dataset: Dataset
    ) -> TensorType:
        """
        Method for calculating new inducing points given a `model` and `dataset`.

        This method is to be implemented by all subclasses of :class:`InducingPointSelector`.

        :param M: Desired number of inducing points.
        :param model: The sparse model.
        :param dataset: The data from the observer.
        :return: The new updated inducing points.
        """
        raise NotImplementedError


class UniformInducingPointSelector(InducingPointSelector[GPflowPredictor]):
    """
    An :class:`InducingPointSelector` that chooses points sampled uniformly across the search space.
    """

    def __init__(self, search_space: SearchSpace, recalc_every_model_update: bool = True):
        """
        :param search_space: The global search space over which the optimization is defined.
        :param recalc_every_model_update: If True then recalculate the inducing points for each
            model update, otherwise just recalculate on the first call.
        """
        super().__init__(recalc_every_model_update)
        self._search_space = search_space

    def _recalculate_inducing_points(
        self, M: int, model: GPflowPredictor, dataset: Dataset
    ) -> TensorType:
        """
        Sample `M` points. If `search_space` is a :class:`Box` then we use a space-filling Sobol
        design to ensure high diversity.

        :param M: Desired number of inducing points.
        :param model: The sparse model .
        :param dataset: The data from the observer.
        :return: The new updated inducing points.
        """

        if isinstance(self._search_space, Box):
            return self._search_space.sample_sobol(M)
        else:
            return self._search_space.sample(M)


class RandomSubSampleInducingPointSelector(InducingPointSelector[GPflowPredictor]):
    """
    An :class:`InducingPointSelector` that chooses points at random from the training data.
    """

    def _recalculate_inducing_points(
        self, M: int, model: GPflowPredictor, dataset: Dataset
    ) -> TensorType:
        """
        Sample `M` points from the training data without replacement. If we require more
        inducing points than training data, then we fill the remaining points with random
        samples across the search space.

        :param M: Desired number of inducing points.
        :param model: The sparse model.
        :param dataset: The data from the observer. Must be populated.
        :return: The new updated inducing points.
        :raise tf.errors.InvalidArgumentError: If ``dataset`` is empty.
        """

        tf.debugging.Assert(len(dataset.query_points) is not None, [tf.constant([])])

        N = tf.shape(dataset.query_points)[0]  # training data size
        shuffled_query_points = tf.random.shuffle(dataset.query_points)  # [N, d]
        sub_sample = shuffled_query_points[: tf.math.minimum(N, M), :]

        if N < M:  # if fewer data than inducing points then sample remaining uniformly
            data_as_discrete_search_space = DiscreteSearchSpace(dataset.query_points)
            convex_hull_of_data = Box(
                lower=data_as_discrete_search_space.lower,
                upper=data_as_discrete_search_space.upper,
            )

            uniform_sampler = UniformInducingPointSelector(convex_hull_of_data)
            uniform_sample = uniform_sampler._recalculate_inducing_points(
                M - N, model, dataset
            )  # [M-N, d]
            sub_sample = tf.concat([sub_sample, uniform_sample], 0)  # [M, d]
        return sub_sample  # [M, d]


class KMeansInducingPointSelector(InducingPointSelector[GPflowPredictor]):
    """
    An :class:`InducingPointSelector` that chooses points as centroids of a K-means clustering
    of the training data.
    """

    def _recalculate_inducing_points(
        self, M: int, model: GPflowPredictor, dataset: Dataset
    ) -> TensorType:
        """
        Calculate `M` centroids from a K-means clustering of the training data.

        If the clustering returns fewer than `M` centroids or if we have fewer than `M` training
        data, then we fill the remaining points with random samples across the search space.

        :param M: Desired number of inducing points.
        :param model: The sparse model.
        :param dataset: The data from the observer. Must be populated.
        :return: The new updated inducing points.
        :raise tf.errors.InvalidArgumentError: If ``dataset`` is empty.
        """

        tf.debugging.Assert(dataset is not None, [tf.constant([])])

        query_points = dataset.query_points  # [N, d]
        N = tf.shape(query_points)[0]

        shuffled_query_points = tf.random.shuffle(query_points)  # [N, d]
        query_points_stds = tf.math.reduce_std(shuffled_query_points, 0)  # [d]

        if (
            tf.math.count_nonzero(query_points_stds, dtype=N.dtype) == N
        ):  # standardize if all stds non zero
            normalize = True
            shuffled_query_points = shuffled_query_points / query_points_stds  # [N, d]
        else:
            normalize = False

        centroids, _ = kmeans(shuffled_query_points, int(tf.math.minimum(M, N)))  # [C, d]

        if normalize:
            centroids *= query_points_stds  # [M, d]

        if len(centroids) < M:  # choose remaining points as random samples
            data_as_discrete_search_space = DiscreteSearchSpace(dataset.query_points)
            convex_hull_of_data = Box(
                lower=data_as_discrete_search_space.lower,
                upper=data_as_discrete_search_space.upper,
            )
            uniform_sampler = UniformInducingPointSelector(convex_hull_of_data)
            extra_centroids = uniform_sampler._recalculate_inducing_points(  # [M-C, d]
                M - len(centroids), model, dataset
            )
            centroids = tf.concat([centroids, extra_centroids], axis=0)  # [M, d]

        return centroids  # [M, d]


class QualityFunction(ABC):
    """
    A :const:`QualityFunction` uses a  `model` to measure the quality of each of
    the `N` query points in the provided `dataset`, returning shape `[N]`.
    """

    @abstractmethod
    def __call__(self, model: GPflowPredictor, dataset: Dataset) -> TensorType:
        """
        Evaluate the quality of the data-points according to the model.
        :param model: The sparse model.
        :param dataset: The data from the observer. Must be populated.
        :return: The quality scores.
        """


class DPPInducingPointSelector(InducingPointSelector[GPflowPredictor]):
    """
    An :class:`InducingPointSelector` that follows :cite:`chen2018fast` to get a greedy appoximation
    to the MAP estimate of the specified Determinantal Point Process (DPP).

    The DPP is defined through its diveristy-quality decomposition, i.e. its similarity kernel
    is just the kernel of the considered model and its quality scores come from the
    provided :class:`QualityFunction`.

    """

    def __init__(self, quality_function: QualityFunction, recalc_every_model_update: bool = True):
        """
        :param quality_function: A function measuring the quality of each candidate inducing point.
        :param recalc_every_model_update: If True then recalculate the inducing points for each
            model update, otherwise just recalculate on the first call.
        """
        super().__init__(recalc_every_model_update)
        self._quality_function = quality_function

    def _recalculate_inducing_points(
        self,
        M: int,
        model: GPflowPredictor,
        dataset: Dataset,
    ) -> TensorType:
        """
        :param M: Desired number of inducing points.
        :param model: The sparse model.
        :param dataset: The data from the observer. Must be populated.
        :return: The new updated inducing points.
        :raise tf.errors.InvalidArgumentError: If ``dataset`` is empty.
        """
        tf.debugging.Assert(dataset is not None, [])
        N = tf.shape(dataset.query_points)[0]

        quality_scores = self._quality_function(model, dataset)

        chosen_inducing_points = greedy_inference_dpp(
            M=tf.minimum(M, N),
            kernel=model.get_kernel(),
            quality_scores=quality_scores,
            dataset=dataset,
        )  # [min(M,N), d]

        if N < M:  # if fewer data than inducing points then sample remaining uniformly
            data_as_discrete_search_space = DiscreteSearchSpace(dataset.query_points)
            convex_hull_of_data = Box(
                lower=data_as_discrete_search_space.lower,
                upper=data_as_discrete_search_space.upper,
            )
            uniform_sampler = UniformInducingPointSelector(convex_hull_of_data)
            uniform_sample = uniform_sampler._recalculate_inducing_points(
                M - N,
                model,
                dataset,
            )  # [M-N, d]
            chosen_inducing_points = tf.concat(
                [chosen_inducing_points, uniform_sample], 0
            )  # [M, d]
        return chosen_inducing_points  # [M, d]


class UnitQualityFunction(QualityFunction):
    """
    A :class:`QualityFunction` where all points are considered equal, i.e. using
    this quality function for inducing point allocation corresponds to allocating
    inducing points with the sole aim of minimizing predictive variance.
    """

    def __call__(self, model: GPflowPredictor, dataset: Dataset) -> TensorType:
        """
        Evaluate the quality of the data-points according to the model.
        :param model: The sparse model.
        :param dataset: The data from the observer. Must be populated.
        :return: The quality scores.
        """

        return tf.ones(tf.shape(dataset.query_points)[0], dtype=tf.float64)  # [N]


class ModelBasedImprovementQualityFunction(QualityFunction):
    """
    A :class:`QualityFunction` where the quality of points are given by their expected
    improvement with respect to a conservative baseline. Expectations are according
    to the model from the previous BO step). See :cite:`moss2023IPA` for details
    and justification.
    """

    def __call__(self, model: GPflowPredictor, dataset: Dataset) -> TensorType:
        """
        Evaluate the quality of the data-points according to the model.
        :param model: The sparse model.
        :param dataset: The data from the observer. Must be populated.
        :return: The quality scores.
        """

        mean, variance = model.predict(dataset.query_points)  # [N, 1], [N, 1]
        baseline = tf.reduce_max(mean)
        normal = tfp.distributions.Normal(mean, tf.sqrt(variance))
        improvement = (baseline - mean) * normal.cdf(baseline) + variance * normal.prob(
            baseline
        )  # [N, 1]
        return improvement[:, 0]  # [N]


class ConditionalVarianceReduction(DPPInducingPointSelector):
    """
    An :class:`InducingPointSelector` that greedily chooses the points with maximal (conditional)
    predictive variance, see :cite:`burt2019rates`.
    """

    def __init__(self, recalc_every_model_update: bool = True):
        """
        :param recalc_every_model_update: If True then recalculate the inducing points for each
            model update, otherwise just recalculate on the first call.
        """

        super().__init__(UnitQualityFunction(), recalc_every_model_update)


class ConditionalImprovementReduction(DPPInducingPointSelector):
    """
    An :class:`InducingPointSelector` that greedily chooses points with large predictive variance
    and that are likely to be in promising regions of the search space, see :cite:`moss2023IPA`.
    """

    def __init__(
        self,
        recalc_every_model_update: bool = True,
    ):
        """
        :param recalc_every_model_update: If True then recalculate the inducing points for each
            model update, otherwise just recalculate on the first call.
        """

        super().__init__(ModelBasedImprovementQualityFunction(), recalc_every_model_update)


def greedy_inference_dpp(
    M: int,
    kernel: gpflow.kernels.Kernel,
    quality_scores: TensorType,
    dataset: Dataset,
) -> TensorType:
    """
    Get a greedy approximation of the MAP estimate of the Determinantal Point Process (DPP)
    over ``dataset`` following the algorithm of :cite:`chen2018fast`. Note that we are using the
    quality-diversity decomposition of a DPP, specifying both a similarity ``kernel``
    and ``quality_scores``.

    :param M: Desired set size.
    :param kernel: The underlying kernel of the DPP.
    :param quality_scores: The quality score of each item in ``dataset``.
    :return: The MAP estimate of the DPP.
    :raise tf.errors.InvalidArgumentError: If ``dataset`` is empty or if the shape of
        ``quality_scores`` does not match that of ``dataset.observations``.
    """
    tf.debugging.Assert(dataset is not None, [])
    tf.debugging.assert_equal(tf.shape(dataset.observations)[0], tf.shape(quality_scores)[0])
    tf.debugging.Assert(len(dataset.query_points) >= M, [])

    chosen_indicies = []  # iteratively store chosen points

    N = tf.shape(dataset.query_points)[0]
    c = tf.zeros((M - 1, N))  # [M-1,N]
    d_squared = kernel.K_diag(dataset.query_points)  # [N]

    scores = d_squared * quality_scores**2  # [N]
    chosen_indicies.append(tf.argmax(scores))  # get first element
    for m in range(M - 1):  # get remaining elements
        ix = tf.cast(chosen_indicies[-1], dtype=tf.int32)  # increment Cholesky with newest point
        newest_point = dataset.query_points[ix]

        d_temp = tf.math.sqrt(d_squared[ix])  # [1]

        L = kernel.K(dataset.query_points, newest_point[None, :])[:, 0]  # [N]
        if m == 0:
            e = L / d_temp
            c = tf.expand_dims(e, 0)  # [1,N]
        else:
            c_temp = c[:, ix : ix + 1]  # [m,1]
            e = (L - tf.matmul(tf.transpose(c_temp), c[:m])) / d_temp  # [N]
            c = tf.concat([c, e], axis=0)  # [m+1, N]
            e = tf.squeeze(e, 0)

        d_squared -= e**2
        d_squared = tf.maximum(d_squared, 1e-50)  # numerical stability

        scores = d_squared * quality_scores**2  # [N]
        chosen_indicies.append(tf.argmax(scores))  # get next element as point with largest score

    return tf.gather(dataset.query_points, chosen_indicies)



# Copyright 2021 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Optional

import tensorflow as tf
import tensorflow_probability as tfp
from gpflow.keras import tf_keras
from typing_extensions import Protocol, runtime_checkable

from ...data import Dataset
from ...space import EncoderFunction
from ...types import TensorType
from ..interfaces import EncodedProbabilisticModel, ProbabilisticModel
from ..optimizer import KerasOptimizer


class KerasPredictor(EncodedProbabilisticModel, ABC):
    """
    This is an interface for trainable wrappers of TensorFlow and Keras neural network models.
    """

    def __init__(
        self,
        optimizer: Optional[KerasOptimizer] = None,
        encoder: EncoderFunction | None = None,
    ):
        """
        :param optimizer: The optimizer wrapper containing the optimizer with which to train the
            model and arguments for the wrapper and the optimizer. The optimizer must
            be an instance of a :class:`~tf.optimizers.Optimizer`. Defaults to
            :class:`~tf.optimizers.Adam` optimizer with default parameters.
        :param encoder: Optional encoder with which to transform query points before
            generating predictions.
        :raise ValueError: If the optimizer is not an instance of :class:`~tf.optimizers.Optimizer`.
        """
        if optimizer is None:
            optimizer = KerasOptimizer(tf_keras.optimizers.Adam())
        self._optimizer = optimizer
        self._encoder = encoder

        if not isinstance(optimizer.optimizer, tf_keras.optimizers.Optimizer):
            raise ValueError(
                f"Optimizer for `KerasPredictor` models must be an instance of a "
                f"`tf.optimizers.Optimizer`, received {type(optimizer.optimizer)} instead."
            )

    @property
    @abstractmethod
    def model(self) -> tf_keras.Model:
        """The compiled Keras model."""
        raise NotImplementedError

    @property
    def optimizer(self) -> KerasOptimizer:
        """The optimizer wrapper for training the model."""
        return self._optimizer

    @property
    def encoder(self) -> EncoderFunction | None:
        return self._encoder

    @encoder.setter
    def encoder(self, encoder: EncoderFunction | None) -> None:
        self._encoder = encoder

    def predict_encoded(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
        return self.model.predict(query_points)

    def sample_encoded(self, query_points: TensorType, num_samples: int) -> TensorType:
        raise NotImplementedError(
            """
            KerasPredictor does not implement sampling. Acquisition
            functions relying on it cannot be used with this class by default. Certain
            types of neural networks might be able to generate samples and
            such subclasses should overwrite this method.
            """
        )

    def log(self, dataset: Optional[Dataset] = None) -> None:
        return


@runtime_checkable
class DeepEnsembleModel(ProbabilisticModel, Protocol):
    """
    This is an interface for deep ensemble type of model, primarily for usage by trajectory
    samplers, to avoid circular imports. These models can act as probabilistic models
    by deriving estimates of epistemic uncertainty from the diversity of predictions made by
    individual models in the ensemble.
    """

    @property
    @abstractmethod
    def ensemble_size(self) -> int:
        """
        Returns the size of the ensemble, that is, the number of base learners or individual
        models in the ensemble.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def num_outputs(self) -> int:
        """
        Returns the number of outputs trained on by each member network.
        """
        raise NotImplementedError

    @abstractmethod
    def ensemble_distributions(
        self, query_points: TensorType
    ) -> tuple[tfp.distributions.Distribution, ...]:
        """
        Return distributions for each member of the ensemble. Type of the output will depend on the
        subclass, it might be a predicted value or a distribution.

        :param query_points: The points at which to return outputs.
        :return: The outputs for the observations at the specified ``query_points`` for each member
            of the ensemble.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def dtype(self) -> tf.DType:
        """The prediction dtype."""



# Copyright 2020 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from pathlib import Path

from setuptools import find_packages, setup

with open("README.md", "r") as file:
    long_description = file.read()

setup(
    name="trieste",
    version=Path("trieste/VERSION").read_text().strip(),
    author="The Trieste contributors",
    author_email="labs@secondmind.ai",
    description="A Bayesian optimization research toolbox built on TensorFlow",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/secondmind-labs/trieste",
    packages=find_packages(include=("trieste*",)),
    package_data={
        "trieste": ["py.typed", "VERSION"],
    },
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires="~=3.9",
    install_requires=[
        "absl-py",
        "dill<0.3.6",
        "gpflow>=2.9.2",
        "gpflux>=0.4.4",
        "numpy",
        "tensorflow>=2.5,<2.17; platform_system!='Darwin' or platform_machine!='arm64'",
        "tensorflow-macos>=2.5,<2.17; platform_system=='Darwin' and platform_machine=='arm64'",
        "tensorflow-probability>=0.13,<0.25",
        "greenlet>=1.1.0",
    ],
    extras_require={
        "plotting": ["seaborn", "plotly", "imageio"],
        "qhsri": ["pymoo", "cvxpy"],
    },
)
